@@ -1,3 +1,7 @@
+#ifdef __CYGWIN__
+#define _GNU_SOURCE 1
+#endif
+
 #include "detect-upc.h"
 
 /*******************************************************************************
@@ -105,6 +109,49 @@ int xform_misc = 0;
 int c_pragma = 0;
 int in_c = 0;
 
+/*******************************************************************************
+ * Normalize an absolute path, returning a new string that caller must free().
+ * Here "normalization" is removal of all instances of "//", "/./" and "/../".
+ * Input must begin with '/' or it will be returned unchanged.
+ * Last path component is left unchanged, even if it is ".", "..", or empty.
+ * In other words: we assume 'in' is the absolute path of a file, but even if
+ * not, we will always return a valid path which corresponds to the input.
+ *******************************************************************************/
+#ifdef __CYGWIN__
+  #include <cygwin/version.h>
+  #if (CYGWIN_VERSION_API_MAJOR > 0) || (CYGWIN_VERSION_API_MINOR >= 213)
+  /* If available, use canonicalize_file_name() which will additionally
+   * resolve all symlinks (which gcc-5 under cygwin treat inconsistently).
+   */
+    #define USE_canonicalize_file_name
+  #endif
+#endif
+
+char *normalize(const char *in) {
+  #ifdef USE_canonicalize_file_name
+    char *out = canonicalize_file_name(in);
+    return out ? out : strdup(in);
+  #else
+    char *p, *q, *r, *out;
+    out = p = strdup(in);
+    while (p[0]=='/') {
+        if (p[1]=='/') {
+            q = p + 1;
+        } else if (p[1]=='.' && p[2]=='/') {
+            q = p + 2;
+        } else if (p[1]=='.' && p[2]=='.' && p[3]=='/') {
+            q = p + 3;
+            while (p!=out && *--p!='/');
+        } else {
+            while (*++p && *p!='/');
+            continue;
+        }
+        for (r = p; (*++r = *++q); );
+    }
+    return out;
+  #endif
+}
+
 /*******************************************************************************
  * The list of known UPC files.
  * Current implemented as a linked list
@@ -134,10 +181,16 @@ static void print_upc_files(void) {
 /* Checks to see if the fname is currently in the upc_files list */
 static int is_upc_file(const char * fname) {
     flist_node * p = upc_files;
+    char * tmpname = normalize(fname);
+    int result = 0;
     for (p = upc_files; p != NULL; p = p->next) {
-   	if (!strcmp(fname, p->filename)) return 1;
+	if (!strcmp(tmpname, p->filename)) {
+	    result = 1;
+	    break;
+	}
     }
-    return 0;
+    free(tmpname);
+    return result;
 }	
 
 /* Adds fname to the upc_files list */
@@ -149,7 +202,7 @@ static flist_node *add_upc_file(const char *fname) {
     }
     p->has_pragma = 0;
     p->next = NULL;
-    p->filename = strdup(fname);
+    p->filename = normalize(fname);
     if (!p->filename) {
 	perror("strdup failed");
 	exit(-1);
@@ -463,7 +516,7 @@ void insert_pragma(char *yytxt)
 	 * by 1 */
     	fprintf(yyout, "%s", remove_ephemera_from_line(yytxt));
 	if (verbose) {
-	    printf("added UPC pragma to %s\n", filename);
+	    printf("added UPC pragma 'upc_code' to %s\n", filename);
 	    fflush(stdout);
 	}
 	in_c = 0;
@@ -478,7 +531,7 @@ void insert_pragma(char *yytxt)
 	     * by 1 */
 	    fprintf(yyout, "%s", remove_ephemera_from_line(yytxt));
 	    if (verbose) {
-	        printf("added UPC pragma to %s\n", filename);
+	        printf("added UPC pragma 'c_code' to %s\n", filename);
 	        fflush(stdout);
 	    }
         }
@@ -626,7 +679,11 @@ void read_dependency(FILE *f)
     	    	    perror("getline failed");
     	    	    exit(-1);
                 }
-    	    	p->includes[i] = strdup(temp);
+#ifdef USE_canonicalize_file_name /* Only on Cygwin 1.7+ for now */
+                p->includes[i] = normalize(temp);
+#else
+                p->includes[i] = strdup(temp);
+#endif
     	    	    if (!p->includes[i]) {
     	    	    	perror("strdup failed");
     	    	    	exit(-1);
