@@ -1,6 +1,4 @@
-/*   $Source: /var/local/cvs/gasnet/extended-ref/gasnet_vis_strided.c,v $
- *     $Date: 2009/05/12 02:16:38 $
- * $Revision: 1.27 $
+/*   $Source: bitbucket.org:berkeleylab/gasnet.git/extended-ref/gasnet_vis_strided.c $
  * Description: GASNet Strided implementation
  * Copyright 2002, Dan Bonachea <bonachea@cs.berkeley.edu>
  * Terms of use are as specified in license.txt
@@ -10,6 +8,12 @@
   #error This file not meant to be compiled directly - included by gasnet_extended_refvis.c
 #endif
 
+/* Clang can be picky */
+#if PLATFORM_COMPILER_CLANG
+  #pragma clang diagnostic push
+  #pragma clang diagnostic ignored "-Wconstant-logical-operand"
+#endif
+
 /*---------------------------------------------------------------------------------*/
 /* ***  Strided *** */
 /*---------------------------------------------------------------------------------*/
@@ -1228,5 +1232,10 @@ extern gasnet_handle_t gasnete_gets(gasnete_synctype_t synctype,
   gasneti_fatalerror("failure in GASNETE_GETS_SELECTOR - should never reach here");
   return GASNET_INVALID_HANDLE; /* avoid warning on MIPSPro */
 }
+
+#if PLATFORM_COMPILER_CLANG
+  #pragma clang diagnostic pop
+#endif
+
 #endif
 /*---------------------------------------------------------------------------------*/
