@@ -47,12 +47,39 @@
   #endif
  #endif
 
+ /* TODO:
+    Below are a couple cases where we want to use the lower portion of the O/S-provided
+    mmap()ed region to create our segment because one must ensure some alignment which
+    is greater than the page size.  Instead of forcing the use of the bottom portion,
+    we *could* change the trimming logic to use a chosen granularity other than that of
+    GASNETI_PAGESIZE (such as the huge page size).
+  */
+
+ #ifdef GASNETI_USE_HUGETLBFS
+  #include <hugetlbfs.h>
+  /* Trim only from top to retain alignment: */
+  #undef GASNETI_USE_HIGHSEGMENT
+  #define GASNETI_USE_HIGHSEGMENT 0
+ #endif
+
+ #if GASNET_PSHM && PLATFORM_ARCH_SPARC
+  /* On SPARC, shared mappings must be L1 data-cache color-aligned.
+     The easiest way to ensure that is to trim only from top to preserve
+     the alignment which the O/S gave to the original (untrimmed) mapping.
+   Possible TODO:
+     Since we only support 64-bit SPARC CPUs the alignment will always be MAX(16K, PAGESIZE).
+     So we could just use that for GASNETI_PAGESIZE.
+   */
+  #undef GASNETI_USE_HIGHSEGMENT
+  #define GASNETI_USE_HIGHSEGMENT 0
+ #endif
+
 #if !HAVE_MMAP
   /* Skip the following platform checks */
 #elif PLATFORM_OS_CYGWIN
   #error mmap not supported on Cygwin - it doesnt work properly
 #elif PLATFORM_ARCH_MIC
-  #define GASNETI_MMAP_FLAGS (MAP_ANONYMOUS | MAP_PRIVATE | MAP_POPULATE | MAP_HUGETLB)
+  #define GASNETI_MMAP_FLAGS (MAP_ANONYMOUS | MAP_PRIVATE | MAP_POPULATE)
 #endif
 
 #ifndef GASNETI_MMAP_FLAGS
@@ -142,6 +169,10 @@ static void *gasneti_mmap_internal(void *segbase, uintptr_t segsize) {
                        (unsigned long)segsize, strerror(mmap_errno));
   }
 
+  if ((ptr != (void*)GASNETI_PAGE_ALIGNDOWN(ptr)) && (ptr != MAP_FAILED)) {
+      gasneti_fatalerror("mmap result "GASNETI_LADDRFMT" is not aligned to GASNET_PAGESIZE %lu (0x%lx)",
+              GASNETI_LADDRSTR(ptr), (unsigned long)GASNET_PAGESIZE, (unsigned long)GASNET_PAGESIZE);
+  }
   if (segbase && ptr == MAP_FAILED) {
       gasneti_fatalerror("mmap fixed failed at "GASNETI_LADDRFMT" for size %lu: %s\n",
 	      GASNETI_LADDRSTR(segbase), (unsigned long)segsize, strerror(mmap_errno));
@@ -237,18 +268,27 @@ static const char *gasneti_pshm_makeunique(const char *unique) {
   static char prefix[] = "/GASNTXXXXXX";
 #if defined(GASNETI_PSHM_FILE) || defined(GASNETI_PSHM_SYSV)
   const char *tmpdir = gasneti_tmpdir();
-  const size_t tmpdir_len = strlen(tmpdir);
-#else
-  const size_t tmpdir_len = 0;
 #endif
+  size_t tmpdir_len = 0;
 #if !defined(GASNETI_PSHM_SYSV)
-  const size_t base_len = tmpdir_len + GASNETI_PSHM_PREFIX_LEN;
+  size_t base_len;
   char *allnames;
 #endif
   int i;
 
   gasneti_assert(strlen(prefix) == GASNETI_PSHM_PREFIX_LEN);
 
+#if defined(GASNETI_PSHM_FILE) && defined(GASNETI_USE_HUGETLBFS)
+  { const char *hugedir = hugetlbfs_find_path();
+    if (hugedir && !access(hugedir, R_OK|W_OK|X_OK)) {
+      tmpdir = hugedir;
+    }
+  }
+#endif
+#if defined(GASNETI_PSHM_FILE) || defined(GASNETI_PSHM_SYSV)
+  tmpdir_len = strlen(tmpdir);
+#endif
+
   if (!unique) { /* We get to pick the unique bits */
 #if defined(GASNETI_PSHM_FILE) || defined(GASNETI_PSHM_SYSV)
     if (gasneti_pshm_mkstemp(prefix, tmpdir)) {
@@ -312,6 +352,7 @@ static const char *gasneti_pshm_makeunique(const char *unique) {
   memcpy(prefix + GASNETI_PSHM_PREFIX_LEN1, unique, GASNETI_PSHM_UNIQUE_LEN);
 
   gasneti_pshmname = (char **)gasneti_malloc((gasneti_pshm_nodes+1)*sizeof(char*));
+  base_len = tmpdir_len + GASNETI_PSHM_PREFIX_LEN;
   allnames = (char *)gasneti_malloc((gasneti_pshm_nodes+1)*(base_len + 4));
 
   for (i = 0; i <= gasneti_pshm_nodes; ++i) {
@@ -341,98 +382,36 @@ static const char *gasneti_pshm_makeunique(const char *unique) {
 #endif /* GASNET_PSHM */
 
 #if defined(GASNETI_USE_HUGETLBFS)
-#include <hugetlbfs.h>
 
-#if defined(HAVE_HUGETLBFS_UNLINKED_FD_FOR_SIZE) && 0 /* Disabled pending further study */
-#define GASNETI_USE_HUGETLBFS_SIZES 1
-#endif
-
-#if defined(GASNETI_USE_HUGETLBFS_SIZES)
-static int compare_long(const void *a_p, const void *b_p) {
-  long a = *(long *)a_p;
-  long b = *(long *)b_p;
-  return (a==b) ? 0 : ((a<b) ? -1 : 1);
-}
-#endif
-
-/* Pick an available hugepage size for mapping of the requested size.
- * With older libhugetlbfs this will always pick the default size.
- * The given addr and size are adjusted for proper alignment.
+/* Apply the default hugepage size for mapping of the requested size.
+ * The given size is adjusted for proper alignment and returned.
  */
-static long pick_pagesz(void **addr_p, uintptr_t *size_p) {
-  uintptr_t size = *size_p;
-
-#if defined(GASNETI_USE_HUGETLBFS_SIZES)
-  /* Currently pick largest size less than the rounded request.
-   * Other possibilities exists, of course.
-   */
-  #define gasneti_unlinked_huge_fd(pgsz) hugetlbfs_unlinked_fd_for_size(pgsz)
-  static long *tbl = NULL;
-  static int count = 0;
-  long pagesz;
-  int i;
-
-  if (!tbl) {
-    long dflt = gethugepagesize();
-    count = gethugepagesizes(NULL, 0);
-    tbl = gasneti_calloc(count+1, sizeof(long)); /* final 0 marks end */
-    gethugepagesizes(tbl, count);
-    qsort(tbl, count, sizeof(long), compare_long);
-    /* remove from consideration any smaller than the default */
-    while (*tbl && (*tbl < dflt)) { ++tbl; --count; }
-  }
-
-  for (i=0; i<count; ++i) {
-    long next = tbl[i+1];
-    pagesz = tbl[i];
-    if (!next || (GASNETI_ALIGNUP(size, pagesz) < next)) break;
-  }
-#elif defined(HAVE_HUGETLBFS_UNLINKED_FD)
-  #define gasneti_unlinked_huge_fd(pgsz) hugetlbfs_unlinked_fd()
+static uintptr_t huge_pagesz(void *addr, uintptr_t size) {
   static long pagesz = 0;
   if (!pagesz) pagesz = gethugepagesize();
-#else
-  #error
-#endif
-
-  if (addr_p) {
-    uintptr_t addr = (uintptr_t)(*addr_p);
-    uintptr_t new_addr = GASNETI_ALIGNDOWN(addr, pagesz);
-    *addr_p = (void*)new_addr;
-    size = GASNETI_ALIGNUP(size + (addr - new_addr), pagesz);
-  } else {
-    size = GASNETI_ALIGNUP(size, pagesz);
-  }
-  *size_p = size;
-
-  return pagesz;
+  gasneti_assert((uintptr_t)addr % pagesz == 0); /* alignment check */
+  return GASNETI_ALIGNUP(size, pagesz);
 }
 
 extern void *gasneti_huge_mmap(void *addr, uintptr_t size) {
-  void *real_addr = addr;
-  GASNETI_UNUSED long pagesz = pick_pagesz(&real_addr, &size);
-  int fd = gasneti_unlinked_huge_fd(pagesz);
+  int fd = hugetlbfs_unlinked_fd();
   const int mmap_flags = MAP_SHARED | (addr ? GASNETI_MMAP_FIXED_FLAG : GASNETI_MMAP_NOTFIXED_FLAG);
-  void *ptr = mmap(real_addr, size, (PROT_READ|PROT_WRITE), mmap_flags, fd, 0);
+  void *ptr = mmap(addr, huge_pagesz(addr, size), (PROT_READ|PROT_WRITE), mmap_flags, fd, 0);
 
+  int save_errno = errno;
   (void) close(fd);
-
-  /* If expanded a fixed mmap then be sure to return the requested fixed addr */
-  if (real_addr != addr) {
-    gasneti_assert(ptr == real_addr);
-    ptr = addr;
-  }
+  errno = save_errno;
 
   return ptr;
 }
 
 extern void gasneti_huge_munmap(void *addr, uintptr_t size) {
-  (void)pick_pagesz(&addr, &size);
-  if (munmap(addr, size) != 0) 
+  if (munmap(addr, huge_pagesz(addr, size)) != 0)
     gasneti_fatalerror("munmap("GASNETI_LADDRFMT",%lu) failed: %s\n",
                        GASNETI_LADDRSTR(addr), (unsigned long)size, strerror(errno));
 }
-#endif
+
+#endif /* defined(GASNETI_USE_HUGETLBFS) */
 
 #if GASNET_PSHM
 
@@ -469,6 +448,9 @@ static void * gasneti_pshm_mmap(int pshm_rank, void *segbase, size_t segsize) {
 
   /* create or open */
   #if defined(GASNETI_PSHM_FILE)
+   #if defined(GASNETI_USE_HUGETLBFS)
+    segsize = huge_pagesz(segbase, segsize);
+   #endif
     fd = open(filename, flags, S_IRUSR | S_IWUSR);
   #elif defined(GASNETI_PSHM_POSIX)
     fd = shm_open(filename, flags, S_IRUSR | S_IWUSR);
@@ -503,7 +485,7 @@ static void * gasneti_pshm_mmap(int pshm_rank, void *segbase, size_t segsize) {
   }
 #elif defined(GASNETI_PSHM_XPMEM)
   if (create) {
-  #if HAVE_HUGETLBFS
+  #if GASNETI_USE_HUGETLBFS
     ptr = gasneti_huge_mmap(segbase, segsize);
   #else
     const int mmap_flags = GASNETI_MAP_ANONYMOUS | MAP_SHARED | \
@@ -511,14 +493,23 @@ static void * gasneti_pshm_mmap(int pshm_rank, void *segbase, size_t segsize) {
     ptr = mmap(segbase, segsize, (PROT_READ|PROT_WRITE), mmap_flags, 0, 0);
   #endif
   } else {
-    gasneti_xpmem_addr_t xa;
-    xa.offset = 0;
-    xa.gasneti_xpmem_apid =
-          gasneti_pshm_apids[pshm_rank] = 
+    gasneti_xpmem_apid_t apid =
+  #if HAVE_XPMEM_MAKE_2
+            xpmem_get_2(gasneti_pshm_segids[pshm_rank], XPMEM_RDWR, XPMEM_PERMIT_MODE, NULL);
+  #else
               xpmem_get(gasneti_pshm_segids[pshm_rank], XPMEM_RDWR, XPMEM_PERMIT_MODE, NULL);
-    if (xa.gasneti_xpmem_apid != (gasneti_xpmem_apid_t)-1) {
+  #endif
+    if (apid != (gasneti_xpmem_apid_t)-1) {
+    #if HAVE_XPMEM_MAKE_2
+      ptr = xpmem_attach_2(apid, 0, segsize, segbase);
+    #else
+      gasneti_xpmem_addr_t xa;
+      xa.offset = 0;
+      xa.gasneti_xpmem_apid = apid;
       ptr = xpmem_attach(xa, segsize, segbase);
+    #endif
     }
+    gasneti_pshm_apids[pshm_rank] = apid;
   }
 #elif defined(GASNETI_PSHM_GHEAP)
   if (create) {
@@ -546,14 +537,8 @@ static void gasneti_pshm_munmap(void *segbase, uintptr_t segsize) {
       gasneti_fatalerror("shmdt("GASNETI_LADDRFMT") failed: %s\n",
 	      GASNETI_LADDRSTR(segbase), strerror(errno));
   }
-#elif defined(GASNETI_PSHM_FILE) || defined(GASNETI_PSHM_POSIX)
+#elif defined(GASNETI_PSHM_FILE) || defined(GASNETI_PSHM_POSIX) || defined(GASNETI_PSHM_XPMEM)
   gasneti_munmap(segbase, segsize);
-#elif defined(GASNETI_PSHM_XPMEM)
- #if HAVE_HUGETLBFS
-  gasneti_huge_munmap(segbase, segsize);
- #else
-  gasneti_munmap(segbase, segsize);
- #endif
 #elif defined(GASNETI_PSHM_GHEAP)
   gasneti_pshm_vfree(segbase);
 #else
@@ -573,8 +558,13 @@ static void gasneti_munmap_remote(gasnet_node_t pshm_rank, void *segbase, uintpt
 #elif defined(GASNETI_PSHM_FILE) || defined(GASNETI_PSHM_POSIX)
   gasneti_munmap(segbase, segsize);
 #elif defined(GASNETI_PSHM_XPMEM)
+ #if HAVE_XPMEM_MAKE_2
+  xpmem_detach_2(segbase, segsize);
+  xpmem_release_2(gasneti_pshm_apids[pshm_rank]);
+ #else
   xpmem_detach(segbase);
   xpmem_release(gasneti_pshm_apids[pshm_rank]);
+ #endif
 #elif defined(GASNETI_PSHM_GHEAP)
  /* Nothing to do here */
 #else
@@ -587,7 +577,12 @@ GASNETI_INLINE(gasneti_export_segment)
 void gasneti_export_segment(void *segbase, uintptr_t segsize) {
 #if defined(GASNETI_PSHM_XPMEM)
   /* Create and supernode-exchange xpmem segment ids */
-  gasneti_xpmem_segid_t segid = xpmem_make(segbase, segsize, XPMEM_PERMIT_MODE, (void *)(uintptr_t)0600);
+  gasneti_xpmem_segid_t segid =
+  #if HAVE_XPMEM_MAKE_2
+          xpmem_make_2(segbase, segsize, XPMEM_PERMIT_MODE, (void *)(uintptr_t)0600);
+  #else
+            xpmem_make(segbase, segsize, XPMEM_PERMIT_MODE, (void *)(uintptr_t)0600);
+  #endif
   if_pf (segid == (gasneti_xpmem_segid_t)(-1)) {
     fprintf(stderr, "xpmem_make() failed:%s\n", strerror(errno));
   }
@@ -599,15 +594,6 @@ void gasneti_export_segment(void *segbase, uintptr_t segsize) {
 #endif
 }
 
-/* Called collectively */
-GASNETI_INLINE(gasneti_unexport_segment)
-void gasneti_unexport_segment(void) {
-#if defined(GASNETI_PSHM_XPMEM)
-  xpmem_remove(gasneti_pshm_segids[gasneti_pshm_mynode]);
-#else
-  /* empty */
-#endif
-}
 /* Helper: destroy the "attach point" */
 static void gasneti_pshm_unlink(int pshm_rank){
 #if defined(GASNETI_PSHM_SYSV)
@@ -733,6 +719,11 @@ static void *gasneti_mmap_shared_internal(int pshmnode, void *segbase, uintptr_t
     }
   }
 
+  if ((ptr != (void*)GASNETI_PAGE_ALIGNDOWN(ptr)) && (ptr != MAP_FAILED)) {
+    gasneti_cleanup_shm();
+    gasneti_fatalerror("mmap result "GASNETI_LADDRFMT" is not aligned to GASNET_PAGESIZE %lu (0x%lx)",
+              GASNETI_LADDRSTR(ptr), (unsigned long)GASNET_PAGESIZE, (unsigned long)GASNET_PAGESIZE);
+  }
 #if !GASNETI_PSHM_MAP_FIXED_IGNORED
   if (segbase && (segbase != ptr) && (ptr != MAP_FAILED)) {
     gasneti_cleanup_shm();
@@ -755,19 +746,18 @@ extern void *gasneti_mmap_shared(uintptr_t segsize) {
   return gasneti_mmap_shared_internal(gasneti_pshm_mynode, NULL, segsize, 1);
 }
 
-extern void *gasneti_mmap_vnet(uintptr_t size, gasneti_bootstrapExchangefn_t exchangefn) {
+extern void *gasneti_mmap_vnet(uintptr_t size, gasneti_bootstrapBroadcastfn_t snodebcastfn) {
   void *ptr = MAP_FAILED;
+  int save_errno = 0;
 
   #if defined(GASNETI_PSHM_SYSV) && PLATFORM_OS_CYGWIN
   /* Cygwin may raise SIGSYS when SysV support is absent.
      This will yield more informative error messages. */
   gasneti_sighandlerfn_t prev_handler = gasneti_reghandler(SIGSYS, SIG_IGN);
-  int save_errno;
   #endif
 
   #if defined(GASNETI_PSHM_FILE) || defined(GASNETI_PSHM_SYSV) || defined(GASNETI_PSHM_POSIX)
   {
-    char (*exchg)[GASNETI_PSHM_UNIQUE_LEN];
     char unique[GASNETI_PSHM_UNIQUE_LEN];
 
     /* First in each supernode creates the segment plus the names/keys and returns the unique identifier */
@@ -775,24 +765,21 @@ extern void *gasneti_mmap_vnet(uintptr_t size, gasneti_bootstrapExchangefn_t exc
       const char *tmp = gasneti_pshm_makeunique(NULL);
       memcpy(unique, tmp, GASNETI_PSHM_UNIQUE_LEN);
       ptr = gasneti_mmap_shared_internal(gasneti_pshm_nodes, NULL, size, 1);
+      save_errno = errno;
     }
 
-    /* Conduit's exchangefn is used as a supernode-scoped bcast to
-     * communicate the unique identifier generated by the firsts */
-    exchg = gasneti_malloc(gasneti_nodes * GASNETI_PSHM_UNIQUE_LEN);
-    (*exchangefn)(unique, GASNETI_PSHM_UNIQUE_LEN, exchg);
-
-    /* Non-first nodes attach and generate the same names/keys from the unique identifier */
+    /* Supernode-scoped bcast communicates the names/keys generated by the firsts */
+    (*snodebcastfn)(unique, GASNETI_PSHM_UNIQUE_LEN, unique, gasneti_pshm_firstnode);
+  
+    /* Non-first nodes attach */
     if (gasneti_pshm_mynode != 0) {
-      (void)gasneti_pshm_makeunique((const char *)(exchg + gasneti_pshm_firstnode));
+      (void)gasneti_pshm_makeunique(unique);
       ptr = gasneti_mmap_shared_internal(gasneti_pshm_nodes, NULL, size, 1);
+      save_errno = errno;
     }
-
-    gasneti_free(exchg);
   }
   #elif defined(GASNETI_PSHM_XPMEM)
   {
-    gasneti_xpmem_segid_t *exchg;
     gasneti_xpmem_segid_t segid = (gasneti_xpmem_segid_t)(-1);
 
     /* Initialization */
@@ -802,24 +789,29 @@ extern void *gasneti_mmap_vnet(uintptr_t size, gasneti_bootstrapExchangefn_t exc
     /* First in each supernode creates the segment */
     if (gasneti_pshm_mynode == 0) {
       ptr = gasneti_mmap_shared_internal(gasneti_pshm_nodes, NULL, size, 1);
+      save_errno = errno;
       if (ptr != MAP_FAILED) {
-        segid = xpmem_make(ptr, size, XPMEM_PERMIT_MODE, (void *)(uintptr_t)0600);
+      #if HAVE_XPMEM_MAKE_2
+        segid = xpmem_make_2(ptr, size, XPMEM_PERMIT_MODE, (void *)(uintptr_t)0600);
+      #else
+        segid =   xpmem_make(ptr, size, XPMEM_PERMIT_MODE, (void *)(uintptr_t)0600);
+      #endif
+        save_errno = errno;
         if_pf (segid == (gasneti_xpmem_segid_t)(-1)) {
           fprintf(stderr, "xpmem_make() failed:%s\n", strerror(errno));
         }
       }
     }
     
-    /* Conduit's exchangefn is used as a supernode-scoped bcast to
-     * communicate the segment identifier generated by the firsts */
-    exchg = gasneti_malloc(gasneti_nodes * sizeof(segid));
-    (*exchangefn)(&segid, sizeof(segid), exchg);
-    gasneti_pshm_segids[gasneti_pshm_nodes] = exchg[gasneti_pshm_firstnode];
-    gasneti_free(exchg);
+    /* Supernode-scoped bcast communicates the segment identifier generated by the firsts */
+    (*snodebcastfn)(&segid, sizeof(segid),
+                    &gasneti_pshm_segids[gasneti_pshm_nodes],
+                    gasneti_pshm_firstnode);
 
     /* Non-first nodes attach */
     if (gasneti_pshm_mynode != 0) {
       ptr = gasneti_mmap_shared_internal(gasneti_pshm_nodes, NULL, size, 1);
+      save_errno = errno;
     }
   }
   #elif defined(GASNETI_PSHM_GHEAP)
@@ -830,22 +822,21 @@ extern void *gasneti_mmap_vnet(uintptr_t size, gasneti_bootstrapExchangefn_t exc
     /* First in each supernode creates the segment */
     if (gasneti_pshm_mynode == 0) {
       ptr = gasneti_mmap_shared_internal(gasneti_pshm_nodes, NULL, size, 1);
+      save_errno = errno;
       if (ptr != MAP_FAILED) {
         memset(ptr, 0, size);
       }
     }
     
-    /* Conduit's exchangefn is used as a supernode-scoped bcast to
-     * communicate the segment addresses alocated by the firsts */
-    { void **exchg = gasneti_malloc(gasneti_nodes * sizeof(ptr));
-      (*exchangefn)(&ptr, sizeof(ptr), exchg);
-      gasneti_pshm_segaddrs[gasneti_pshm_nodes] = exchg[gasneti_pshm_firstnode];
-      gasneti_free(exchg);
-    }
+    /* Supernode-scoped bcast communicates the segment addresses allocated by the firsts */
+    (*snodebcastfn)(&ptr, sizeof(ptr),
+                    &gasneti_pshm_segaddrs[gasneti_pshm_nodes],
+                    gasneti_pshm_firstnode);
 
     /* Non-first nodes attach */
     if (gasneti_pshm_mynode != 0) {
       ptr = gasneti_mmap_shared_internal(gasneti_pshm_nodes, NULL, size, 1);
+      save_errno = errno;
     }
   }
   #else
@@ -853,11 +844,12 @@ extern void *gasneti_mmap_vnet(uintptr_t size, gasneti_bootstrapExchangefn_t exc
   #endif
 
   #if defined(GASNETI_PSHM_SYSV) && PLATFORM_OS_CYGWIN
-  save_errno = errno;
   gasneti_reghandler(SIGSYS, prev_handler);
-  errno = save_errno;
   #endif
 
+  /* restore the pertinent errno, if any */
+  errno = save_errno;
+
   return (ptr == MAP_FAILED) ? NULL : ptr;
 }
 extern void gasneti_unlink_vnet(void) {
@@ -871,9 +863,13 @@ extern void gasneti_munmap(void *segbase, uintptr_t segsize) {
   gasneti_tick_t t1, t2;
   gasneti_assert(segsize > 0);
   t1 = gasneti_ticks_now();
+  #if defined(GASNETI_USE_HUGETLBFS)
+    gasneti_huge_munmap(segbase, segsize);
+  #else
     if (munmap(segbase, segsize) != 0) 
       gasneti_fatalerror("munmap("GASNETI_LADDRFMT",%lu) failed: %s\n",
 	      GASNETI_LADDRSTR(segbase), (unsigned long)segsize, strerror(errno));
+  #endif
   t2 = gasneti_ticks_now();
 
   GASNETI_TRACE_PRINTF(D,("munmap("GASNETI_LADDRFMT", %lu): %.3fus\n", 
@@ -1081,15 +1077,14 @@ static gasnet_seginfo_t gasneti_segment = {0,0}; /* local segment info */
 static uintptr_t gasneti_myheapend = 0; /* top of my malloc heap */
 static uintptr_t gasneti_maxheapend = 0; /* top of max malloc heap */
 static uintptr_t gasneti_maxbase = 0; /* start of segment overlap region */
-#if GASNET_PSHM
-static gasnet_seginfo_t *gasneti_remote_segments;
-#endif /* GASNET_PSHM */
 #endif /* GASNETI_MMAP_OR_PSHM */
 
 typedef struct {
   gasnet_seginfo_t seginfo;
-  uintptr_t heapend;
-  uintptr_t segsize_request; /* during attach only */
+  union {
+    uintptr_t heapend; /* during init only */
+    uintptr_t segsize_request; /* during attach only */
+  } u;
 } gasneti_segexch_t;
 static gasneti_segexch_t *gasneti_segexch = NULL; /* exchanged segment information */
 
@@ -1114,6 +1109,7 @@ uintptr_t gasneti_mmapLimit(uintptr_t localLimit, uint64_t sharedLimit,
                             gasneti_bootstrapBarrierfn_t barrierfn) {
   int i;
   uintptr_t maxsz;
+  const gasnet_node_t local_count = gasneti_myhost.node_count;
 
 #if GASNET_PSHM
   gasneti_pshm_cs_enter();
@@ -1152,25 +1148,25 @@ uintptr_t gasneti_mmapLimit(uintptr_t localLimit, uint64_t sharedLimit,
   maxsz = MIN(maxsz, localLimit);
 
   /* Coordinate the search IFF there are any shared nodes. */
-  if (gasneti_nodemap_global_count != gasneti_nodes) {
+  if (gasneti_myhost.grp_count != gasneti_nodes) {
     uintptr_t *sz_exchg = gasneti_malloc(gasneti_nodes * sizeof(uintptr_t));
     gasnet_seginfo_t se = {0,0};
 
     /* Ensure our probe will not collectively exceed the shareLimit, if any. */
-    if ((sharedLimit != (uint64_t)-1) && (gasneti_nodemap_local_count > 1)) {
+    if ((sharedLimit != (uint64_t)-1) && (local_count > 1)) {
 #if SIZEOF_VOID_P != 8
        /* Skip MIN() on overflow */
-       if ((sharedLimit / gasneti_nodemap_local_count) < (uint64_t)(uintptr_t)(-1))
+       if ((sharedLimit / local_count) < (uint64_t)(uintptr_t)(-1))
 #endif
-       { uintptr_t tmp = sharedLimit / gasneti_nodemap_local_count;
+       { uintptr_t tmp = sharedLimit / local_count;
          maxsz = MIN(maxsz, tmp);
        }
     }
 
-    /* Allow each node to probe SEQUENTIALLY, and then collect the results */
+    /* Allow each node in a given host to probe SEQUENTIALLY, and then collect the results */
     maxsz = GASNETI_PAGE_ALIGNDOWN(maxsz);
 #if GASNET_PSHM
-    if (maxsz) {
+    if (maxsz && (gasneti_myhost.grp_count == gasneti_mysupernode.grp_count)) { /* host==supernode */
       for (i = 0; i < gasneti_nodemap_local_count; ++i) {
         if (i == gasneti_nodemap_local_rank) {
           se = _gasneti_mmap_segment_search_inner(maxsz);
@@ -1180,46 +1176,38 @@ uintptr_t gasneti_mmapLimit(uintptr_t localLimit, uint64_t sharedLimit,
         gasneti_pshmnet_bootstrapBroadcast(gasneti_request_pshmnet, &maxsz, sizeof(uintptr_t), &maxsz, i);
         sz_exchg[gasneti_nodemap_local[i]] = maxsz;
       }
-    }
-#else
+    } else
+#endif
     if (maxsz) {
-      /* Find widest supernode */
+      /* Find widest host */
       gasnet_node_t rounds = 0;
       {
-        gasnet_node_t *tmp = gasneti_calloc(gasneti_nodemap_global_count, sizeof(gasnet_node_t));
+        const gasnet_node_t num_hosts = gasneti_myhost.grp_count;
+        gasnet_node_t *tmp = gasneti_calloc(num_hosts, sizeof(gasnet_node_t));
         for (i = 0; i < gasneti_nodes; ++i) {
-          const gasnet_node_t supernode = gasneti_nodeinfo[i].supernode;
-          tmp[supernode] += 1;
-          rounds = MAX(rounds, tmp[supernode]);
+          const gasnet_node_t host = gasneti_nodeinfo[i].host;
+          gasneti_assert(host < num_hosts);
+          tmp[host] += 1;
+          rounds = MAX(rounds, tmp[host]);
         }
         gasneti_free(tmp);
       }
 
       for (i = 0; i < rounds; ++i) {
-        if (i == gasneti_nodemap_local_rank) {
+        if (i == gasneti_myhost.node_rank) {
           se = _gasneti_mmap_segment_search_inner(maxsz);
         }
         (*barrierfn)();
       }
     }
     (*exchangefn)(&se.size, sizeof(uintptr_t), sz_exchg);
-#endif
 
-    /* Compute the supernode-local mean */
-    { uint64_t sum;
-      gasnet_node_t first, j;
-
-      first = gasneti_nodemap[gasneti_mynode];
-      sum = sz_exchg[first];
-      j = 1;
-
-      for (i = (first + 1); j < gasneti_nodemap_local_count; ++i) {
-        if (gasneti_nodemap[i] == first) {
-          sum += sz_exchg[i];
-          j += 1;
-        }
+    /* Compute the host-local mean */
+    { uint64_t sum = 0;
+      for (i = 0; i < local_count; ++i) {
+        sum += sz_exchg[gasneti_myhost.nodes[i]];
       }
-      maxsz = sum / gasneti_nodemap_local_count;
+      maxsz = sum / local_count;
       maxsz = GASNETI_PAGE_ALIGNDOWN(maxsz);
 
 #if GASNET_PSHM
@@ -1340,28 +1328,11 @@ void gasneti_segmentInit(uintptr_t localSegmentLimit,
     if (gasneti_myheapend == (uintptr_t)-1) gasneti_fatalerror("Failed to sbrk(0):%s",strerror(errno));
     gasneti_myheapend = GASNETI_PAGE_ALIGNUP(gasneti_myheapend);
   #endif
-    se.heapend = gasneti_myheapend;
-    se.segsize_request = 0;
+    se.u.heapend = gasneti_myheapend;
 
     /* gather the sbrk info and mmap segment location */
     (*exchangefn)(&se, sizeof(gasneti_segexch_t), gasneti_segexch);
 
-#if GASNET_PSHM
-    /* Map the remote shared segments */
-    gasneti_export_segment(gasneti_segment.addr, gasneti_segment.size);
-    gasneti_remote_segments = gasneti_malloc(gasneti_pshm_nodes*sizeof(gasnet_seginfo_t));
-    for(i=0; i<gasneti_pshm_nodes; i++){
-        const gasnet_node_t j = gasneti_nodemap_local[i];
-        if (j == gasneti_mynode) continue;
-
-        gasneti_remote_segments[i].addr = gasneti_mmap_remote_shared(NULL,gasneti_segexch[j].seginfo.size,i);
-        gasneti_remote_segments[i].size = gasneti_segexch[j].seginfo.size;
-    }
-
-    /* Unlink the shared segments to prevent leaks (they are recreated in segmentAttach) */
-    gasneti_unlink_segments();
-#endif
- 
     /* compute bounding-box of segment location */
     { uintptr_t maxbase = 0;
       uintptr_t maxsize = 0;
@@ -1371,8 +1342,8 @@ void gasneti_segmentInit(uintptr_t localSegmentLimit,
       char segstats[255];
       /* compute various stats across nodes */
       for (i=0;i < gasneti_nodes; i++) {
-        if (gasneti_segexch[i].heapend > maxheapend)
-          maxheapend = gasneti_segexch[i].heapend;
+        if (gasneti_segexch[i].u.heapend > maxheapend)
+          maxheapend = gasneti_segexch[i].u.heapend;
         if (((uintptr_t)gasneti_segexch[i].seginfo.addr) > maxbase)
           maxbase = (uintptr_t)gasneti_segexch[i].seginfo.addr;
         if (gasneti_segexch[i].seginfo.size > maxsize)
@@ -1434,7 +1405,7 @@ void gasneti_segmentInit(uintptr_t localSegmentLimit,
                       GASNETI_LADDRSTR(gasneti_segexch[i].seginfo.addr), 
                       GASNETI_LADDRSTR(((uintptr_t)gasneti_segexch[i].seginfo.addr)+gasneti_segexch[i].seginfo.size), 
                       (unsigned long)gasneti_segexch[i].seginfo.size,
-                      GASNETI_LADDRSTR(gasneti_segexch[i].heapend));
+                      GASNETI_LADDRSTR(gasneti_segexch[i].u.heapend));
               fflush(stderr);
             }
           }
@@ -1475,28 +1446,22 @@ void gasneti_segmentInit(uintptr_t localSegmentLimit,
 }
 
 /* ------------------------------------------------------------------------------------ */
-static
-void gasneti_segmentAttachLocal(uintptr_t segsize, uintptr_t minheapoffset,
+
+void gasneti_segmentAttach(uintptr_t segsize, uintptr_t minheapoffset,
                            gasnet_seginfo_t *seginfo,
                            gasneti_bootstrapExchangefn_t exchangefn) {
   void *segbase = NULL;
-  uintptr_t bias;
   gasneti_assert(seginfo);
   gasneti_assert(exchangefn);
   gasneti_assert(gasneti_segexch);
   gasneti_memcheck(gasneti_segexch);
 
-  #ifndef GASNETI_SEGMENT_DISALIGN_BIAS
-    #if GASNET_DEBUG && !GASNET_ALIGNED_SEGMENTS && (GASNET_PAGESIZE < 1024*1024)
-      /* force segment disalignment for debugging purposes */
-      #define GASNETI_SEGMENT_DISALIGN_BIAS GASNET_PAGESIZE
-    #else
-      #define GASNETI_SEGMENT_DISALIGN_BIAS 0
-    #endif
+  #if GASNET_PSHM
+    /* Avoid leaking shared memory files in case of non-collective exit between init/attach */
+    gasneti_pshm_cs_enter();
+    gasneti_pshmnet_bootstrapBarrier();
   #endif
 
-  bias = GASNETI_SEGMENT_DISALIGN_BIAS * (gasneti_mynode%2);
-
   #ifdef GASNETI_MMAP_OR_PSHM
   { /* TODO: this assumes heap grows up */
     uintptr_t topofheap;
@@ -1512,15 +1477,14 @@ void gasneti_segmentAttachLocal(uintptr_t segsize, uintptr_t minheapoffset,
 
           /* gather the segsize info again */
           se.seginfo = gasneti_segment;
-          se.heapend = gasneti_myheapend;
-          se.segsize_request = segsize;
+          se.u.segsize_request = segsize;
           (*exchangefn)(&se, sizeof(gasneti_segexch_t), gasneti_segexch);
 
           for (i=0;i<gasneti_nodes;i++) {
             uintptr_t segstart = 
                 ((uintptr_t)gasneti_segexch[i].seginfo.addr + gasneti_segexch[i].seginfo.size) - 
-                 gasneti_segexch[i].segsize_request;
-            /*gasneti_assert(gasneti_segexch[i].segsize_request >= 0); True by typing */
+                 gasneti_segexch[i].u.segsize_request;
+            /*gasneti_assert(gasneti_segexch[i].u.segsize_request >= 0); True by typing */
             gasneti_assert(segstart >= gasneti_maxbase);
             if (segstart < minsegstart) minsegstart = segstart;
           }
@@ -1536,18 +1500,8 @@ void gasneti_segmentAttachLocal(uintptr_t segsize, uintptr_t minheapoffset,
       #if GASNETI_USE_HIGHSEGMENT
         segbase = (void *)((uintptr_t)gasneti_segment.addr + 
                            gasneti_segment.size - segsize);
-        if ((intptr_t)(gasneti_segment.size - segsize) >= bias) {
-          segbase = (void *)((uintptr_t)segbase - bias);
-        } else {
-          segbase = (void *)((uintptr_t)segbase + bias);
-          segsize -= bias;
-        }
       #else
         segbase = gasneti_segment.addr;
-        if (gasneti_segment.size > bias) {
-          segbase = (void *)((uintptr_t)segbase + bias);
-          segsize = MIN(segsize,gasneti_segment.size - bias);
-        }
       #endif
     #endif /* GASNETI_ALIGNED_SEGMENTS */
 
@@ -1587,18 +1541,19 @@ void gasneti_segmentAttachLocal(uintptr_t segsize, uintptr_t minheapoffset,
         gasneti_do_mmap_fixed(segbase, segsize);
       }
     }
+    gasneti_free(gasneti_segexch);
+    gasneti_segexch = NULL;
   }
   #else /* !GASNETI_MMAP_OR_PSHM */
     /* for the T3E, and other platforms which don't support mmap */
-    segbase = gasneti_malloc_allowfail(segsize + GASNET_PAGESIZE + bias);
+    segbase = gasneti_malloc_allowfail(segsize + GASNET_PAGESIZE);
     while (!segbase) {
       segsize = GASNETI_PAGE_ALIGNDOWN(segsize/2);
       if (segsize == 0) break; 
-      segbase = gasneti_malloc_allowfail(segsize + GASNET_PAGESIZE + bias);
+      segbase = gasneti_malloc_allowfail(segsize + GASNET_PAGESIZE);
     }
     if (segbase) {
       segbase = (void *)GASNETI_PAGE_ALIGNUP(segbase);
-      segbase = (void *)(((uintptr_t)segbase)+bias);
     }
   #endif /* GASNETI_MMAP_OR_PSHM */
   gasneti_assert(((uintptr_t)segbase) % GASNET_PAGESIZE == 0);
@@ -1619,131 +1574,39 @@ void gasneti_segmentAttachLocal(uintptr_t segsize, uintptr_t minheapoffset,
       }
     }
   #endif
-}
 
-#if GASNET_PSHM
-static uintptr_t
-gasneti_AttachRemote(const gasnet_node_t pshm_node,
-                     uintptr_t minheapoffset, gasnet_seginfo_t *seginfo) {
-  void *segbase = NULL;
-  uintptr_t topofheap;
-  gasnet_node_t node = gasneti_nodemap_local[pshm_node];
-  uintptr_t segsize = seginfo[node].size;
-
-  gasneti_assert(seginfo);
-  gasneti_assert(node != gasneti_mynode);
-
-    topofheap = gasneti_myheapend;
-    
-    #if GASNETI_USE_HIGHSEGMENT
-      segbase = (void *)((uintptr_t)gasneti_remote_segments[pshm_node].addr + 
-                         gasneti_remote_segments[pshm_node].size - segsize);
-    #else
-      segbase = gasneti_remote_segments[pshm_node].addr;
-    #endif
-    
-    if (segsize == 0) {
-      gasneti_munmap_remote(pshm_node, gasneti_remote_segments[pshm_node].addr,
-                                       gasneti_remote_segments[pshm_node].size);
-      segbase = NULL; 
-    }
-    else {
-      if ((((uintptr_t)segbase + segsize) > topofheap) &&
-        (topofheap + minheapoffset > (uintptr_t)segbase)) {
-        uintptr_t maxsegsz;
-        void *endofseg = (void *)((uintptr_t)gasneti_remote_segments[pshm_node].addr + gasneti_remote_segments[pshm_node].size);
-        segbase = (void *)(topofheap + minheapoffset);
-        if (segbase >= endofseg) {
-          fprintf(stderr, "ERROR: minheapoffset too large to accomodate a segment\n");
-          return 1; /* Failure */
-        }
-        maxsegsz = (uintptr_t)endofseg - (uintptr_t)segbase;
-        if (segsize > maxsegsz) {
-          GASNETI_TRACE_PRINTF(I, ("WARNING: gasneti_segmentAttach() reducing requested segsize (%lu=>%lu) to accomodate minheapoffset",
-          (unsigned long)segsize, (unsigned long)maxsegsz));
-          segsize = maxsegsz;
-        }
-      }
-
-    #if defined(GASNETI_PSHM_XPMEM)
-      /* ALWAYS need to reattach */
-    #else
-      /* Avoid reattach when possible */
-      if (gasneti_remote_segments[pshm_node].addr != segbase ||
-          gasneti_remote_segments[pshm_node].size != segsize)
-    #endif
-      {
-        gasneti_assert(segbase >= gasneti_remote_segments[pshm_node].addr &&
-               (uintptr_t)segbase + segsize <= (uintptr_t)gasneti_remote_segments[pshm_node].addr + gasneti_remote_segments[pshm_node].size);
-        gasneti_munmap_remote(pshm_node, gasneti_remote_segments[pshm_node].addr,
-                                         gasneti_remote_segments[pshm_node].size);
-#if GASNETI_PSHM_MAP_FIXED_IGNORED
-        segbase =
-#endif
-        gasneti_mmap_remote_shared(segbase, segsize, pshm_node);
-        gasneti_remote_segments[pshm_node].addr = segbase;
-        gasneti_remote_segments[pshm_node].size = segsize;
-      }
-    }
-    gasneti_assert(((uintptr_t)segbase) % GASNET_PAGESIZE == 0);
-    gasneti_assert(segsize % GASNET_PAGESIZE == 0);
-    GASNETI_TRACE_PRINTF(C, ("Final segment: segbase="GASNETI_LADDRFMT"  segsize=%lu",
-    GASNETI_LADDRSTR(segbase), (unsigned long)segsize));
-
-    gasneti_assert(gasneti_nodeinfo);
-    gasneti_nodeinfo[node].offset = (uintptr_t)segbase - (uintptr_t)seginfo[node].addr;
-  
-    if (segsize < seginfo[node].size){
-        fprintf(stderr,"ERROR: Not enough memory! Process %d tried mapping %lu bytes, but only %lu bytes available.\n",
-                           (int)gasneti_mynode, (unsigned long)seginfo[node].size, (unsigned long)segsize);
-        return segsize; /* Failure */
-    }
-
-    return 0;
-}
-#endif /* GASNET_PSHM */
-
-void gasneti_segmentAttach(uintptr_t segsize, uintptr_t minheapoffset,
-                           gasnet_seginfo_t *seginfo,
-                           gasneti_bootstrapExchangefn_t exchangefn) {
-#if GASNET_PSHM
+  #if GASNET_PSHM
+  { /* Map the remote shared segments */
     int i;
-    uintptr_t ar = 0; /* results of gasneti_AttachRemote */
-
-    gasneti_pshm_cs_enter();
-
-    /* Avoid leaking shared memory files in case of non-collective exit between init/attach */
-    gasneti_pshmnet_bootstrapBarrier();
-
-    gasneti_unexport_segment();
-#endif
-
-    gasneti_segmentAttachLocal(segsize, minheapoffset, seginfo, exchangefn);
-    (*exchangefn)(&gasneti_segment, sizeof(gasnet_seginfo_t), seginfo);
-
-#if GASNET_PSHM
-    gasneti_export_segment(gasneti_segment.addr, gasneti_segment.size);
 
     gasneti_nodeinfo[gasneti_mynode].offset = 0;
-    for(i=0; i<gasneti_pshm_nodes; i++){
-      if (i == gasneti_pshm_mynode) continue;
-      ar = gasneti_AttachRemote(i, minheapoffset, seginfo);
-      if (ar) break;
+    gasneti_export_segment(gasneti_segment.addr, gasneti_segment.size);
+    for (i = 0; i < gasneti_pshm_nodes; i++){
+        if (i != gasneti_pshm_mynode) {
+            const gasnet_node_t node = gasneti_nodemap_local[i];
+            const uintptr_t size = seginfo[node].size;
+            void *segbase = gasneti_mmap_remote_shared(NULL, size, i);
+
+            gasneti_assert(((uintptr_t)segbase) % GASNET_PAGESIZE == 0);
+            gasneti_assert(size % GASNET_PAGESIZE == 0);
+
+            if (((uintptr_t)segbase >= gasneti_myheapend) &&
+                ((uintptr_t)segbase <  gasneti_myheapend + minheapoffset))
+              gasneti_fatalerror("minheapoffset too large to accomodate a remote segment");
+
+            gasneti_nodeinfo[node].offset = (uintptr_t)segbase - (uintptr_t)seginfo[node].addr;
+
+            GASNETI_TRACE_PRINTF(C, ("Remote segment %d: segbase="GASNETI_LADDRFMT"  segsize=%lu",
+                                     (int)node, GASNETI_LADDRSTR(segbase), (unsigned long)size));
+        }
     }
 
-  gasneti_free(gasneti_remote_segments);
-
-  gasneti_pshmnet_bootstrapBarrier();
-  gasneti_cleanup_shm();
-  if (ar) {
-    gasneti_fatalerror("Failed to attach one or more remote segments");
+    gasneti_pshmnet_bootstrapBarrier();
+    gasneti_cleanup_shm();
+    gasneti_pshm_cs_leave();
   }
-
-  gasneti_pshm_cs_leave();
-#endif /* GASNET_PSHM */
-  gasneti_free(gasneti_segexch);
-  gasneti_segexch = NULL;
-} 
+  #endif /* GASNET_PSHM */
+}
 #endif /* !GASNET_SEGMENT_EVERYTHING */
 
 /* Used to pass the nodemap information to the client
@@ -1895,10 +1758,6 @@ static gasneti_auxseg_request_t *gasneti_auxseg_alignedsz = NULL;
 static uintptr_t gasneti_auxseg_sz = 0;
 static uintptr_t gasneti_auxseg_client_request_sz = 0;
 
-#if GASNETI_SEGMENT_DISALIGN_BIAS
-  GASNETI_IDENT(gasneti_disalign_auxseg_IdentString, "$GASNetAuxSeg_disalign: "_STRINGIFY(GASNETI_SEGMENT_DISALIGN_BIAS)" $");
-#endif
-
 #if GASNET_DEBUG
   /* spawner hint of our auxseg requirements */
   #define GASNETI_AUXSEG_DUMMY_SZ    463
@@ -1964,10 +1823,6 @@ void gasneti_auxseg_init(void) {
   #if GASNET_SEGMENT_EVERYTHING
     GASNETI_TRACE_PRINTF(C, ("gasneti_auxseg_init(): gasneti_auxseg_sz = %lu", (unsigned long)gasneti_auxseg_sz));
   #else
-    #if GASNETI_SEGMENT_DISALIGN_BIAS
-      gasneti_auxseg_sz += GASNETI_SEGMENT_DISALIGN_BIAS;
-    #endif
-
     /* TODO: implement request downsizing down to minsz */
     if (gasneti_auxseg_sz >= gasneti_MaxGlobalSegmentSize)
       gasneti_fatalerror("GASNet internal auxseg size (%llu bytes) exceeds available segment size (%llu bytes)",
@@ -2036,9 +1891,6 @@ uintptr_t gasneti_auxseg_preattach(uintptr_t client_request_sz) {
     result = 0;
   }
   #else
-    #if GASNETI_SEGMENT_DISALIGN_BIAS
-      gasneti_auxseg_sz -= GASNETI_SEGMENT_DISALIGN_BIAS;
-    #endif
     gasneti_assert(client_request_sz % GASNET_PAGESIZE == 0);
     gasneti_auxseg_client_request_sz = client_request_sz;
     #if GASNETI_AUXSEG_PRESERVE_POW2_FULLSEGSZ
