@@ -1,8 +1,6 @@
-/*   $Source: /var/local/cvs/gasnet/gasnet_pshm.c,v $
- *     $Date: 2013/08/08 16:52:01 $
- * $Revision: 1.61 $
+/*   $Source: bitbucket.org:berkeleylab/gasnet.git/gasnet_pshm.c $
  * Description: GASNet infrastructure for shared memory communications
- * Copyright 2012, E. O. Lawrence Berekely National Laboratory
+ * Copyright 2012, The Regents of the University of California
  * Terms of use are as specified in license.txt
  */
 
@@ -56,53 +54,37 @@ static struct gasneti_pshm_info {
 #define pshmnet_get_struct_addr_from_field_addr(structname, fieldname, fieldaddr) \
         ((structname*)(((uintptr_t)fieldaddr) - offsetof(structname,fieldname)))
 
-void *gasneti_pshm_init(gasneti_bootstrapExchangefn_t exchangefn, size_t aux_sz) {
+void *gasneti_pshm_init(gasneti_bootstrapBroadcastfn_t snodebcastfn, size_t aux_sz) {
   size_t vnetsz, mmapsz;
   int discontig = 0;
-  gasneti_pshm_rank_t *pshm_max_nodes;
   gasnet_node_t i;
 #if !GASNET_CONDUIT_SMP
   gasnet_node_t j;
 #endif
 
-  gasneti_assert(exchangefn != NULL);  /* NULL exchangefn no longer supported */
+  gasneti_assert(snodebcastfn != NULL);  /* NULL snodebcastfn no longer supported */
 
-  /* Testing if the number of PSHM nodes is always smaller than GASNETI_PSHM_MAX_NODES */
-  pshm_max_nodes = gasneti_calloc(gasneti_nodes, sizeof(gasneti_pshm_rank_t));
-  for(i=0; i<gasneti_nodes; i++){
-    /* Note combination of post-increment and == guard against overflow */
-    if ((pshm_max_nodes[gasneti_nodemap[i]]++) == GASNETI_PSHM_MAX_NODES){
-      if (gasneti_mynode == gasneti_nodemap[i]){
-        gasneti_fatalerror("PSHM nodes requested on node '%s' exceeds maximum (%d)\n", 
-                        gasneti_gethostname(), GASNETI_PSHM_MAX_NODES);
-      } else {
-        gasneti_fatalerror("PSHM nodes requested on some node exceeds maximum (%d)\n", 
-                        GASNETI_PSHM_MAX_NODES);
-      }
-    }
-  }
-  gasneti_free(pshm_max_nodes);
+  gasneti_assert_always(gasneti_nodemap_local_count <= GASNETI_PSHM_MAX_NODES);
     
   gasneti_pshm_nodes = gasneti_nodemap_local_count;
-  gasneti_pshm_firstnode = gasneti_nodemap[gasneti_mynode];
   gasneti_pshm_mynode = gasneti_nodemap_local_rank;
+  gasneti_pshm_firstnode = gasneti_nodemap_local[0];
 
 #if GASNET_CONDUIT_SMP
   gasneti_assert(gasneti_pshm_nodes == gasneti_nodes);
-  gasneti_assert(gasneti_pshm_firstnode == 0);
   gasneti_assert(gasneti_pshm_mynode == gasneti_mynode);
+  gasneti_assert(gasneti_pshm_firstnode == 0);
 #else
-  gasneti_assert(gasneti_nodemap[0] == 0);
-  for (i=1; i<gasneti_nodes; ++i) {
-    /* Determine if supernode members are numbered contiguously */
-    if (gasneti_nodemap[i-1] > gasneti_nodemap[i]) {
+  /* Determine if local supernode members are numbered contiguously */
+  for (i = 1; i < gasneti_nodemap_local_count; ++i) {
+    if (gasneti_nodemap_local[i] != gasneti_pshm_firstnode + i) {
       discontig = 1;
       break;
     }
   }
 #endif
 
-  gasneti_assert(gasneti_pshm_supernodes > 0);
+  gasneti_assert(gasneti_nodemap_global_count > 0);
 
   /* compute size of vnet shared memory region */
   vnetsz = gasneti_pshmnet_memory_needed(gasneti_pshm_nodes); 
@@ -110,7 +92,7 @@ void *gasneti_pshm_init(gasneti_bootstrapExchangefn_t exchangefn, size_t aux_sz)
   { /* gasneti_pshm_info contains multiple variable-length arrays in the same space */
     size_t info_sz;
     /* space for gasneti_pshm_firsts: */
-    info_sz = gasneti_pshm_supernodes * sizeof(gasnet_node_t);
+    info_sz = gasneti_nodemap_global_count * sizeof(gasnet_node_t);
     /* optional space for gasneti_pshm_rankmap: */
     if (discontig) {
       info_sz = GASNETI_ALIGNUP(info_sz, sizeof(gasneti_pshm_rank_t));
@@ -130,7 +112,8 @@ void *gasneti_pshm_init(gasneti_bootstrapExchangefn_t exchangefn, size_t aux_sz)
 
   /* setup vnet shared memory region for AM infrastructure and supernode barrier.
    */
-  gasnetc_pshmnet_region = gasneti_mmap_vnet(mmapsz, exchangefn);
+  gasnetc_pshmnet_region = gasneti_mmap_vnet(mmapsz, snodebcastfn);
+  gasneti_assert_always((((uintptr_t)gasnetc_pshmnet_region) % GASNETI_PSHMNET_PAGESIZE) == 0);
   if (gasnetc_pshmnet_region == NULL) {
     const int save_errno = errno;
     char buf[16];
@@ -167,9 +150,9 @@ void *gasneti_pshm_init(gasneti_bootstrapExchangefn_t exchangefn, size_t aux_sz)
   gasneti_pshmnet_bootstrapBarrier();
   {
     uintptr_t addr = (uintptr_t)&gasneti_pshm_info->early_barrier;
-    /* gasneti_pshm_firsts, an array of gasneti_pshm_supernodes*sizeof(gasnet_node_t): */
+    /* gasneti_pshm_firsts, an array of gasneti_nodemap_global_count*sizeof(gasnet_node_t): */
     gasneti_pshm_firsts = (gasnet_node_t *)addr;
-    addr += gasneti_pshm_supernodes * sizeof(gasnet_node_t);
+    addr += gasneti_nodemap_global_count * sizeof(gasnet_node_t);
     /* optional rankmap: */
     if (discontig) {
       addr = GASNETI_ALIGNUP(addr, sizeof(gasneti_pshm_rank_t));
@@ -190,14 +173,14 @@ void *gasneti_pshm_init(gasneti_bootstrapExchangefn_t exchangefn, size_t aux_sz)
     if (gasneti_nodemap[i] == i) {
       if (!gasneti_pshm_mynode) gasneti_pshm_firsts[j] = i;
       j += 1;
-      gasneti_assert(j <= gasneti_pshm_supernodes);
+      gasneti_assert(j <= gasneti_nodemap_global_count);
     }
   }
 #endif
 #if GASNET_DEBUG
   /* Validate gasneti_pshm_firsts[] */
   if (gasneti_pshm_mynode == 0) {
-    for (i=0; i<gasneti_pshm_supernodes; ++i) {
+    for (i=0; i<gasneti_nodemap_global_count; ++i) {
       gasneti_assert(!i || (gasneti_pshm_firsts[i] > gasneti_pshm_firsts[i-1]));
       gasneti_assert(gasneti_nodemap[gasneti_pshm_firsts[i]] == gasneti_pshm_firsts[i]);
     }
@@ -354,20 +337,8 @@ typedef union {
                   gasneti_atomic_set((_t),0,0)
   #define gasneti_pshmnet_tail_cas(_t,_o,_n)\
                   gasneti_atomic_compare_and_swap((_t),(_o),(_n),0)
- #if GASNETI_HAVE_ATOMIC_SWAP
   #define gasneti_pshmnet_tail_swap(_t,_v)\
                   gasneti_atomic_swap((_t),(_v),GASNETI_ATOMIC_REL)
- #else
-  GASNETI_INLINE(gasneti_pshmnet_tail_swap)
-  gasneti_atomic_val_t gasneti_pshmnet_tail_swap(gasneti_atomic_t *t, gasneti_atomic_val_t val) {
-    gasneti_atomic_val_t old_val;
-    gasneti_local_wmb();
-    do {
-      old_val = gasneti_atomic_read(t, 0);
-    } while (!gasneti_atomic_compare_and_swap(t, old_val, val, 0));
-    return old_val;
-  }
- #endif /* HAVE_SWAP */
 #elif defined(GASNETI_HAVE_ATOMIC_ADD_SUB)
   typedef struct {
     gasneti_atomic_t last_ticket;
@@ -593,22 +564,18 @@ size_t gasneti_pshmnet_memory_needed(gasneti_pshm_rank_t nodes)
 /* Initializes the pshmnet region. Called from each node twice: 
  * to initialize pshmnet_request and pshmnet_reply */
 gasneti_pshmnet_t *
-gasneti_pshmnet_init(void *start, size_t nbytes, gasneti_pshm_rank_t pshmnodes)
+gasneti_pshmnet_init(void *region, size_t regionlen, gasneti_pshm_rank_t pshmnodes)
 {
   gasneti_pshmnet_t *vnet;
-  size_t szonce, szpernode, regionlen;
-  void *region, *myregion;
+  size_t szonce, szpernode;
+  void *myregion;
 
   /* make sure that our max buffer size fits all possible AMs */
   gasneti_assert(sizeof(gasneti_AMPSHM_maxmsg_t) <= GASNETI_PSHMNET_MAX_PAYLOAD);
 
-  region = start;
-  region = (void *)round_up_to_pshmpage(region);
-
   szpernode = gasneti_pshmnet_memory_needed_pernode(pshmnodes);
   szonce = gasneti_pshmnet_memory_needed_once(pshmnodes);
 
-  regionlen = nbytes - ( ((uintptr_t)region)-((uintptr_t)start));
   if (regionlen < (szonce + szpernode * pshmnodes))
     gasneti_fatalerror("Internal error: not enough memory for pshmnet: \n"
                        " given %lu effective bytes, but need %lu", 
@@ -687,7 +654,7 @@ int gasneti_pshmnet_queue_peek(const gasneti_pshmnet_queue_t * const q)
 int gasneti_pshmnet_recv(gasneti_pshmnet_t *vnet, void **pbuf, size_t *psize, 
                          gasneti_pshm_rank_t *pfrom)
 {
-  gasneti_atomic_val_t head, next;
+  gasneti_atomic_val_t head;
   gasneti_pshmnet_payload_t *p = NULL;
   gasneti_pshmnet_queue_t *q = vnet->my_queue;
 
@@ -702,16 +669,19 @@ int gasneti_pshmnet_recv(gasneti_pshmnet_t *vnet, void **pbuf, size_t *psize,
       q->head = 0;
     }
     if_pt (head) {
+      register gasneti_atomic_val_t next;
       p = gasneti_pshm_addr(head);
-      next = p->next;
-      q->shead = next;
-      if (!next && !gasneti_pshmnet_tail_cas(&q->tail, head, 0)) {
-        while (0 == (next = p->next)) GASNETI_WAITHOOK(); /* waituntil() has excess RMB */
-        q->shead = next;
-      }
-    #if !GASNET_PAR
       gasneti_local_rmb(); /* ACQ */
-    #endif
+      /* NOTE: Unlike in the Nemesis paper, we loop on *both* p->next and
+       * cas(tail) to allow weaker memory models which may reorder their
+       * respective reads/writes.  This is preferred over adding any memory
+       * fence(s) to the race-free case.
+       */
+      while (GASNETT_PREDICT_FALSE(0 == (next = p->next)) &&
+             GASNETT_PREDICT_FALSE(!gasneti_pshmnet_tail_cas(&q->tail, head, 0))) {
+        GASNETI_WAITHOOK(); /* waituntil() has excess RMB */
+      }
+      q->shead = next;
     }
 #if GASNET_PAR
     gasneti_mutex_unlock(&vnet->lock);
