@@ -211,6 +211,7 @@ extern void (*gasneti_bootstrapBarrier_p)(void);
 extern void (*gasneti_bootstrapExchange_p)(void *src, size_t len, void *dest);
 extern void (*gasneti_bootstrapAlltoall_p)(void *src, size_t len, void *dest);
 extern void (*gasneti_bootstrapBroadcast_p)(void *src, size_t len, void *dest, int rootnode);
+extern void (*gasneti_bootstrapSNodeCast_p)(void *src, size_t len, void *dest, int rootnode);
 extern void (*gasneti_bootstrapCleanup_p)(void);
 #define gasneti_bootstrapFini           (*gasneti_bootstrapFini_p)
 #define gasneti_bootstrapAbort          (*gasneti_bootstrapAbort_p)
@@ -218,6 +219,7 @@ extern void (*gasneti_bootstrapCleanup_p)(void);
 #define gasneti_bootstrapExchange       (*gasneti_bootstrapExchange_p)
 #define gasneti_bootstrapAlltoall       (*gasneti_bootstrapAlltoall_p)
 #define gasneti_bootstrapBroadcast      (*gasneti_bootstrapBroadcast_p)
+#define gasneti_bootstrapSNodeBroadcast (*gasneti_bootstrapSNodeCast_p)
 #define gasneti_bootstrapCleanup        (*gasneti_bootstrapCleanup_p)
 
 #define MXM_MAX_ADDR_LEN 1024
@@ -246,7 +248,9 @@ typedef struct _gasnet_mxm_module {
     uint8_t                    * need_fence;
     uint8_t                      strict_api;
     mxm_mq_h                     mxm_mq;
+#if MXM_API < MXM_VERSION(2,0)
     size_t                       zcopy_thresh;/* MXM threshold for using zcopy*/
+#endif
     size_t                       max_am_med;  /* max size for medium AM */
     gasnetc_memreg_t           * reg;         /* registered segments */
     int                          am_max_depth; 
