@@ -7,6 +7,9 @@
 #include <gasnet_internal.h>
 #include <gasnet_handler.h>
 #include <gasnet_core_internal.h>
+#if GASNET_BLCR
+#include <gasnet_blcr.h>
+#endif
 
 #include <amudp_spmd.h>
 
@@ -26,15 +29,16 @@ static void gasnetc_on_exit(int, void*);
 static void gasnetc_atexit(void);
 #endif
 
+static uint64_t gasnetc_networkpid;
 eb_t gasnetc_bundle;
 ep_t gasnetc_endpoint;
 
 gasneti_mutex_t gasnetc_AMlock = GASNETI_MUTEX_INITIALIZER; /*  protect access to AMUDP */
 volatile int gasnetc_AMLockYield = 0;
 
-#if GASNET_PSHM
+#ifdef GASNETC_MAX_NUMHANDLERS
   gasneti_handler_fn_t gasnetc_handler[GASNETC_MAX_NUMHANDLERS]; /* shadow handler table */
-#endif /* GASNET_PSHM */
+#endif /* GASNETC_MAX_NUMHANDLERS */
 
 #if GASNET_TRACE
   extern void gasnetc_enteringHandler_hook(amudp_category_t cat, int isReq, int handlerId, void *token, 
@@ -45,8 +49,6 @@ volatile int gasnetc_AMLockYield = 0;
 #if defined(GASNET_CSPAWN_CMD)
   #define GASNETC_DEFAULT_SPAWNFN C
   GASNETI_IDENT(gasnetc_IdentString_Default_CSpawnCommand, "$GASNetCSpawnCommand: " GASNET_CSPAWN_CMD " $");
-#elif defined(REXEC)
-  #define GASNETC_DEFAULT_SPAWNFN R
 #else /* AMUDP implicit ssh startup */
   #define GASNETC_DEFAULT_SPAWNFN S
 #endif
@@ -83,6 +85,17 @@ void gasnetc_bootstrapExchange(void *src, size_t len, void *dest) {
   if_pf (retval) gasneti_fatalerror("failure in gasnetc_bootstrapExchange()");
 }
 
+#if GASNET_PSHM /* Used only in call to gasneti_pshm_init() */
+/* Naive (poorly scaling) "reference" implementation via gasnetc_bootstrapExchange() */
+static void gasnetc_bootstrapSNodeBroadcast(void *src, size_t len, void *dest, int rootnode) {
+  void *tmp = gasneti_malloc(len * gasneti_nodes);
+  gasneti_assert(NULL != src);
+  gasnetc_bootstrapExchange(src, len, tmp);
+  memcpy(dest, (void*)((uintptr_t)tmp + (len * rootnode)), len);
+  gasneti_free(tmp);
+}
+#endif
+
 #define INITERR(type, reason) do {                                      \
    if (gasneti_VerboseErrors) {                                         \
      fprintf(stderr, "GASNet initialization encountered an error: %s\n" \
@@ -100,7 +113,7 @@ static int gasnetc_init(int *argc, char ***argv) {
   gasnetc_check_config();
 
   /* --------- begin Master code ------------ */
-  if (!AMUDP_SPMDIsWorker(*argv)) { 
+  if (!AMUDP_SPMDIsWorker(argv?*argv:NULL)) {
     /* assume we're an implicit master 
        (we don't currently support explicit workers spawned 
         without using the AMUDP SPMD API)   
@@ -110,6 +123,10 @@ static int gasnetc_init(int *argc, char ***argv) {
     char spawnfn;
     amudp_spawnfn_t fp = (amudp_spawnfn_t)NULL;
 
+    if (!argv) {
+      gasneti_fatalerror("implicit-master without argv not supported - use amudprun");
+    }
+
     /* pretend we're node 0, for purposes of verbose env reporting */
     gasneti_init_done = 1;
     gasneti_mynode = 0;
@@ -206,7 +223,7 @@ static int gasnetc_init(int *argc, char ***argv) {
     /*  perform job spawn */
     retval = AMUDP_SPMDStartup(argc, argv, 
       0, 0, NULL, /* dummies */
-      NULL, &gasnetc_bundle, &gasnetc_endpoint);
+      &gasnetc_networkpid, &gasnetc_bundle, &gasnetc_endpoint);
     if (retval != AM_OK) INITERR(RESOURCE, "slave AMUDP_SPMDStartup() failed");
     gasneti_init_done = 1; /* enable early to allow tracing */
 
@@ -232,7 +249,7 @@ static int gasnetc_init(int *argc, char ***argv) {
     gasneti_nodemapInit(&gasnetc_bootstrapExchange, NULL, 0, 0);
 
     #if GASNET_PSHM
-      gasneti_pshm_init(&gasnetc_bootstrapExchange, 0);
+      gasneti_pshm_init(&gasnetc_bootstrapSNodeBroadcast, 0);
     #endif
 
     #if GASNET_SEGMENT_FAST || GASNET_SEGMENT_LARGE
@@ -252,6 +269,11 @@ static int gasnetc_init(int *argc, char ***argv) {
       #error Bad segment config
     #endif
 
+    #if GASNET_BLCR
+      gasneti_checkpoint_guid = gasnetc_networkpid;
+      gasneti_checkpoint_init(NULL);
+    #endif
+
   AMUNLOCK();
 
   gasneti_auxseg_init(); /* adjust max seg values based on auxseg */
@@ -321,8 +343,8 @@ static int gasnetc_reghandlers(gasnet_handlerentry_t *table, int numentries,
     /* register the handler */
     if (AM_SetHandler(gasnetc_endpoint, (handler_t)newindex, table[i].fnptr) != AM_OK) 
       GASNETI_RETURN_ERRR(RESOURCE, "AM_SetHandler() failed while registering handlers");
-#if GASNET_PSHM
-    /* Maintain a shadown handler table for AMPSHM */
+#ifdef GASNETC_MAX_NUMHANDLERS
+    /* Maintain a shadow handler table */
     gasnetc_handler[(gasnet_handler_t)newindex] = (gasneti_handler_fn_t)table[i].fnptr;
 #endif
 
@@ -372,8 +394,8 @@ extern int gasnetc_attach(gasnet_handlerentry_t *table, int numentries,
 
     /* ------------------------------------------------------------------------------------ */
     /*  register handlers */
-#if GASNET_PSHM
-    /* Initialize AMPSHM's shadow handler table */
+#ifdef GASNETC_MAX_NUMHANDLERS
+    /* Initialize shadow handler table */
     { int i;
       for (i=0; i<GASNETC_MAX_NUMHANDLERS; i++)
           gasnetc_handler[i]=(gasneti_handler_fn_t)&gasneti_defaultAMHandler;
@@ -826,7 +848,7 @@ extern int gasnetc_AMReplyLongM(
   No-interrupt sections
   =====================
   This section is only required for conduits that may use interrupt-based handler dispatch
-  See the GASNet spec and http://www.cs.berkeley.edu/~bonachea/upc/gasnet.html for
+  See the GASNet spec and http://gasnet.lbl.gov/dist/docs/gasnet.html for
     philosophy and hints on efficiently implementing no-interrupt sections
   Note: the extended-ref implementation provides a thread-specific void* within the 
     gasnete_threaddata_t data structure which is reserved for use by the core 
@@ -983,6 +1005,67 @@ extern int  gasnetc_hsl_trylock(gasnet_hsl_t *hsl) {
   }
 #endif
 
+/* ------------------------------------------------------------------------------------ */
+/*
+  Checkpoint/restart
+  ==================
+  thin wrappers around AMUDP-level support
+*/
+
+#if GASNET_BLCR
+
+/* NON-collective checkpoint request */
+int gasnet_checkpoint(const char *dir) {
+  int i, rc;
+
+  gasneti_flush_streams();
+
+  AMLOCK();
+  rc = AMUDP_SPMDCheckpoint(&gasnetc_bundle, &gasnetc_endpoint, dir);
+  if (rc > 0) {
+    /* Handlers */
+    for (i=0; i<GASNETC_MAX_NUMHANDLERS; i++) {
+      if (gasnetc_handler[i] != (gasneti_handler_fn_t)&gasneti_defaultAMHandler) {
+        AM_SetHandler(gasnetc_endpoint, (handler_t)i, gasnetc_handler[i]);
+        /* BLCR-TODO: error-checking */
+      }
+    }
+
+    /* Segment */
+    i = AM_SetSeg(gasnetc_endpoint,
+                  gasneti_seginfo[gasneti_mynode].addr,
+                  gasneti_seginfo[gasneti_mynode].size);
+    /* BLCR-TODO: error-checking */
+
+    #if GASNET_TRACE
+      if (GASNETI_TRACE_ENABLED(A))
+        GASNETI_AM_SAFE(AMUDP_SetHandlerCallbacks(gasnetc_endpoint,
+          gasnetc_enteringHandler_hook, gasnetc_leavingHandler_hook));
+    #endif
+  }
+  AMUNLOCK();
+
+#if GASNET_DEBUG_VERBOSE
+  fprintf(stderr, "Node %d %s checkpoint\n", gasneti_mynode, rc?"restart from":"continue after");
+#endif
+
+  return rc;
+}
+
+/* Collective checkpoint request */
+
+int gasnet_all_checkpoint(const char *dir_arg) {
+  const char *dir;
+  int rc;
+  dir = gasneti_checkpoint_dir(dir_arg);
+  gasnet_barrier(0, GASNET_BARRIERFLAG_ANONYMOUS);
+  rc = gasnet_checkpoint(dir);
+  if (!dir_arg) gasneti_free((void*)dir);
+  gasnet_barrier(0, GASNET_BARRIERFLAG_ANONYMOUS);
+  return rc;
+}
+
+#endif /* GASNET_BLCR */
 /* ------------------------------------------------------------------------------------ */
 /*
   Private Handlers:
