@@ -43,7 +43,7 @@ my @may_defs;
 my @isys;
 if ($cc_family =~ m/^xlc$/i) { # IBM VisualAge C
   my $xlcentry=`$cc -v $tmpfile 2>&1`;
-  if ($xlcentry =~ m/(xlcentry|xlCcpp)\((.*?)\)/) {
+  if ($xlcentry =~ m/(xl[cC]entry|xlCcpp)\((.*?)\)/) {
     my @xlcentryopts = split(/,/,$2);
     verbose("got xlcentry:\n".join("\n  ",@xlcentryopts)."\n");
     OPT: for my $opt (@xlcentryopts) {
@@ -197,6 +197,7 @@ for my $line (@lines) {
 if (!defined $gcc_defines{"__STDC__"}) { # gcc does define __STDC__, but -dM lies about it
   $gcc_defines{"__STDC__"} = 1;
 }
+delete $gcc_defines{"__STDC_VERSION__"}; # We cannot safely undef this
 
 for my $def (sort keys %gcc_defines) {
     open(TESTFILE, ">$tmpfile") or die "Can't open '$tmpfile' for write\n";
@@ -262,7 +263,7 @@ if ($cc_family =~ m/^xlc$/i) { # IBM VisualAge C
 } elsif ($cc_family =~ m/^Sun$/i) { # Sun CC
   # list from "Sun C User's Guide - Sun WorkShop 6 update 2, July 2001, rev A"
   # Updates from Studio 12.2 compiler man page (esp. for linux and amd64 additions)
-  @cc_maybes = (qw<__STDC__ sun unix sparc i386 __amd64 __x86_64
+  @cc_maybes = (qw<sun unix sparc i386 __amd64 __x86_64
                    __sun __unix __sparc __i386 __SUNPRO_C
                    __SunOS linux __linux __linux__ __gnu__linux
                    __BUILTIN_VA_ARG_INCR __SVR4 __sparcv9
@@ -280,7 +281,7 @@ if ($cc_family =~ m/^xlc$/i) { # IBM VisualAge C
   # I can't find a documented list of predefined PGI macros anywhere...
   # these are all experimentally determined or taken from headers
   # 2011.02.22: Updated using -dM support in pgcc-11.0
-  @cc_maybes = (qw<__STDC__ __STDC_VERSION__ _OPENMP
+  @cc_maybes = (qw<_OPENMP
                    __PGI __PGIC__ __PGIC_MINOR__ __PGIC_PATCHLEVEL__ 
 		   _PGI_NOBUILTINS __PGI_LIBC __PGI_GNUC_VA_LIST PGI_DEBUG>);
   # Can try -dM for REAL list from recent pgcc
@@ -288,8 +289,7 @@ if ($cc_family =~ m/^xlc$/i) { # IBM VisualAge C
 } elsif ($cc_family =~ m/^Compaq$/i) { # Compaq C
   # list from "Tru64 UNIX Compaq C Language Reference Manual, AA-RH9NE-TE, September 2002"
   # cc -E also lets through some cc-specific #pragmas, which gcc discards, but that should be ok
-  @cc_maybes = (qw< __STDC__  __STDC_VERSION__ __STDC_HOSTED__ __STDC_ISO_10646__
-  		    __DECC __DECC_MODE_RELAXED __DECC_MODE_STRICT __DECC_MODE_COMMON __DECC_VER 
+  @cc_maybes = (qw< __DECC __DECC_MODE_RELAXED __DECC_MODE_STRICT __DECC_MODE_COMMON __DECC_VER 
 		    LANGUAGE_C __LANGUAGE_C__
 		    __ALPHA __Alpha_AXP __INITIAL_POINTER_SIZE __PRAGMA_ENVIRONMENT 
                     unix __unix__ __osf SYSTYPE_BSD _SYSTYPE_BSD __alpha
@@ -299,8 +299,8 @@ if ($cc_family =~ m/^xlc$/i) { # IBM VisualAge C
 		  >);
 } elsif ($cc_family =~ m/^Cray$/i) { # Cray C on XT, XE, XK and XC series
   # list from Chapter 9 of "Cray C and C++ Reference Manual" S-2179-81, September 2012
-  # plus some guesswork: __STDC_VERSION__ __STDC_HOSTED__ _OPENMP
-  @cc_maybes = (qw< __STDC__ __STDC_VERSION__ __STDC_HOSTED__ _OPENMP
+  # plus some guesswork: __STDC_HOSTED__ _OPENMP
+  @cc_maybes = (qw< _OPENMP
 		    __linux __linux__ linux __gnu_linux__
 		    _ADDR64 __LITTLE_ENDIAN__ _LITTLE_ENDIAN
 		    _MAXVL_8 _MAXVL_16 _MAXVL_32 _MAXVL_64 _MAXVL_128
@@ -310,10 +310,17 @@ if ($cc_family =~ m/^xlc$/i) { # IBM VisualAge C
   $dM_cmd = "$cc -Wp,-list_final_macros $tmpfile 2>/dev/null >$outfile";
 } else { die "unknown --cc-family=$cc_family\n"; }
 
+# C99 spec-defined preprocessor symbols we always look for
+push @cc_maybes, (qw< __STDC__ __STDC_HOSTED__ __STDC_IEC_559__ >);
+
+# push @cc_maybes, "__STDC_VERSION__";         # BUPC requires C99
+# push @cc_maybes, "__STDC_ISO_10646__";       # BUPC does not support wchar_t
+# push @cc_maybes, "__STDC_IEC_559_COMPLEX__"; # BUPC does not support  _Complex
+
 # Try to get exact list from cc's equivalent of -dM, if any.  
 if (defined($dM_cmd) && !system($dM_cmd) && open(OUTFILE, $outfile)) {
   my @special_defs = qw/__FILE__ __BASE_FILE__ __LINE__ __TIME__ __DATE__
-                        __FUNCTION__ __PRETTY_FUNCTION__/;
+                        __FUNCTION__ __PRETTY_FUNCTION__ __STDC_VERSION__/;
   while (<OUTFILE>) {
     if (/\s*#define\s+(\w+)/) {
       push @cc_maybes, $1
@@ -332,6 +339,7 @@ my @cc_check = grep { my $d = $_;
                        !grep /$d/, @gcc_keep; } @cc_maybes;
 
 for my $def (@cc_check) {
+    next if ($def eq '__STDC_VERSION__'); # Cannot safely redefine
     open(TESTFILE, ">$tmpfile") or die "Can't open '$tmpfile' for write\n";
     print TESTFILE <<EOF;
 #ifndef $def
@@ -370,15 +378,15 @@ if ($cc_family =~ m/^xlc$/i) { # IBM VisualAge C
   # the _Builtin keyword appears in several of xlc 7.0's headers, notably in stdarg.h
   $extradefs .= ' -D_Builtin=';
 
-  # Problem: xlc for Linux defines the following macros:
+  # Problem: xlc for Linux sometimes defines the following macros:
   # __GNUC__ __GNUC_MINOR__
   # only while processing header files that are in a directory specified by 
   # -qgcc_c_stdinc or -qgcc_cpp_stdinc (ie the system directories)
   # we have no way to emulate this behavior, so we have to approximate it with a -D option (overkill)
   # otherwise PPC/Linux sys/types.h will think we lack the long long type and drop int64_t
   if (`uname -s 2> /dev/null` =~ /Linux/) {
-    $extradefs .= ' -D__GNUC__=3';
-    $extradefs .= ' -D__GNUC_MINOR__=3';
+    $extradefs .= ' -D__GNUC__=3'        unless grep {$_ eq '__GNUC__'} @gcc_keep;
+    $extradefs .= ' -D__GNUC_MINOR__=3'  unless grep {$_ eq '__GNUC_MINOR__'} @gcc_keep;
   }
 }
 
