@@ -14,14 +14,21 @@
     fi
     echo "  Run:"
     if [ -f "$dir/run.rpt" ] ; then
-      echo "    SUCCESS:  `cat $dir/run_*.log | grep -cw SUCCESS`"
-      echo "      KNOWN:  `cat $dir/run_*.log | grep -cw KNOWN`"
-      echo "       TIME:  `cat $dir/run_*.log | grep -cw TIME/NEW`"
-      echo "        NEW:  `cat $dir/run_*.log | grep -c '[^E]/NEW'`"
+      s=`cat $dir/run_*.log | grep -cw SUCCESS`
+      k=`cat $dir/run_*.log | grep -cw KNOWN`
+      n=`cat $dir/run_*.log | grep -c  /NEW`
+      t=`cat $dir/run_*.log | grep -cw TIME/NEW`
+      m=`cat $dir/run_*.log | grep -cw MEM/NEW`
+      echo "    SUCCESS:  $s"
+      echo "      KNOWN:  $k"
+      echo "        NEW:  `expr $n - $t - $m`"
+      [ $t != 0 ] && echo "       TIME:  $t"
+      [ $m != 0 ] && echo "        MEM:  $m"
     else
       echo "   No status to report"
     fi
-    y=`find $dir -name 'runlist_*' -print | egrep 'runlist_[0-9]+_[0-9]+$'`
+    y=`find $dir -name 'runlist_*' -print | egrep 'runlist_[0-9]+_[0-9]+$' | \
+       while read f; do test -e ${f}-complete || echo $f; done`
     i=`echo $y | wc -w`
     if [ "$i" -gt 0 ]; then
       echo "   + $i runlist(s) pending (`cat $y 2>/dev/null | wc -l` entries)"
