@@ -10,7 +10,7 @@ use POSIX ":sys_wait_h";
 use Text::ParseWords;
 use Config;
 
-my $warning_blacklist = '^'.
+my $warning_blacklist = '^('.
   '(.*?rmsapi: Error:.*?)|'.
   '(.*?libmsql: Socket read.*?)|'.
   '(.*?Retrying allocation .... press control-C to terminate.*?)|'.
@@ -22,7 +22,8 @@ my $warning_blacklist = '^'.
   '(.*?ENV parameter:.*?)|'.
   '(.*?FREEZE_ON_ERROR.*?)|'.
   '(.*?using fake authentication data for X11 forwarding.*?)|'.
-  '()$';
+  '(.*?most recent RAS event text: DDR (Maintenance )?Correctable Error Summary.*?)|'.
+  ')$';
 
 my $mydir = $0;
 $mydir =~ s@/[^/]*$@@;
