@@ -31,7 +31,7 @@
 #else
 //# error "CONFIGDIR undefined."
 #define debug_init()
-#define debug_step
+#define debug_step ((void)0)
 #define debug(x) x
 #endif
 
