@@ -2,5 +2,5 @@
 #define _SHARED_H
 #include <cyc_a.h>
 #include <upc.h>
-extern shared int blar;
+shared int blar;
 #endif
