@@ -670,25 +670,6 @@ _upcr_alloc(size_t nbytes UPCRI_PT_ARG)
   return retval;
 }
 
-/* LEGACY: to be removed */
-upcr_shared_ptr_t _upcr_local_alloc(size_t nblocks, size_t blocksz UPCRI_PT_ARG)
-{
-  static int did_warn = 0;
-  if_pf (!did_warn) {
-    did_warn = 1;
-    if (!upcr_getenv("UPC_NO_DEPRECATED")) {
-      char loc[500] = { '\0', };
-      upcri_append_srcloc(loc, sizeof(loc));
-      upcri_warn("upc_local_alloc() was removed in UPC 1.3.  "
-                 "Use upc_alloc() instead.  "
-                 "This message appears at most once per thread and can be "
-                 "silenced by setting environment variable UPC_NO_DEPRECATED.%s", loc);
-    }
-  }
-
-  return _upcr_alloc((nblocks * blocksz) UPCRI_PT_PASS);
-}
-
 void upcri_expand_local_SRequest(gasnet_token_t token, 
 				 void *neededlheapsz,
 				 void *request_sz,
