@@ -26,7 +26,7 @@ extern int _upcri_pthread_barrier(void);
 
 #else
 
-#define upcri_pthread_barrier() ((void)0)
+#define upcri_pthread_barrier() (1)
 
 #endif  /* UPCRI_UPC_PTHREADS */
 
