@@ -42,6 +42,6 @@
 #endif
 
 /* Intrepid GCC-UPC doesn't use pthreads */
-#if UPCRI_USING_GCCUPC
+#if UPCRI_USING_GCCUPC || UPCRI_USING_CUPC
 # include "upcr_gupc.c"
 #endif
