@@ -288,7 +288,11 @@ extern uintptr_t	UPCRL_default_shared_offset;
  * 'last.o' object files that are then passed to the linker as the first and
  * last objects on the linker command line).
  */
-#if UPCR_USING_LINKADDRS || UPCRI_USING_GCCUPC
+#if UPCRI_USING_GCCUPC || UPCRI_USING_CUPC
+  #define UPCRL_shared_begin __upc_shared_start
+  #define UPCRL_shared_end __upc_shared_end
+#endif
+#if UPCRI_HAVE_LINKER_SECTION
   extern char UPCRL_shared_begin[0x10000];
   extern char UPCRL_shared_end[1];
 #endif
