@@ -156,8 +156,8 @@ size_t upcri_memvecmove(int isget, void *_localbuf, size_t localsz,
       upcri_assert(handlepos < handlecount);
       upcri_assert(localpos + len <= localsz);
       handles[handlepos++] = (isget ?
-        bupc_memget_async(localbuf+localpos, addr, len) :
-        bupc_memput_async(addr, localbuf+localpos, len) );
+        _upcr_memget_nb(localbuf+localpos, addr, len UPCRI_PT_PASS) :
+        _upcr_memput_nb(addr, localbuf+localpos, len UPCRI_PT_PASS) );
       localpos += len;
     } else {
       /* one or more full blocks, possibly followed by a trailing partial block 
@@ -198,8 +198,8 @@ size_t upcri_memvecmove(int isget, void *_localbuf, size_t localsz,
             upcr_shared_ptr_t remoteaddr = upcr_add_shared(addr, 1, blockid*blocksize, blocksize);
             upcri_assert(handlepos < handlecount);
             handles[handlepos++] = (isget ?
-              bupc_memget_async(localbuf+localpos, remoteaddr, blocksize) :
-              bupc_memput_async(remoteaddr, localbuf+localpos, blocksize) );
+              _upcr_memget_nb(localbuf+localpos, remoteaddr, blocksize UPCRI_PT_PASS) :
+              _upcr_memput_nb(remoteaddr, localbuf+localpos, blocksize UPCRI_PT_PASS) );
             localpos += blocksize;
           }
         }
@@ -209,8 +209,8 @@ size_t upcri_memvecmove(int isget, void *_localbuf, size_t localsz,
         upcri_assert(handlepos < handlecount);
         upcri_assert(localpos + partiallen <= localsz);
         handles[handlepos++] = (isget ?
-          bupc_memget_async(localbuf+localpos, remoteaddr, partiallen) :
-          bupc_memput_async(remoteaddr, localbuf+localpos, partiallen) );
+          _upcr_memget_nb(localbuf+localpos, remoteaddr, partiallen UPCRI_PT_PASS) :
+          _upcr_memput_nb(remoteaddr, localbuf+localpos, partiallen UPCRI_PT_PASS) );
         localpos += partiallen;
       }
     }
@@ -1223,8 +1223,8 @@ static void upcri_process_mbox(upcr_pshared_ptr_t fd, int islocal, int isread UP
         }
         lmbox_ref[threadid].buf1 = upcri_checkmalloc(UPCRI_IO_PAD+mboxlen);
         lmbox_ref[threadid].memcpy_handle = 
-          bupc_memget_async((void *)UPCRI_ALIGNUP(lmbox_ref[threadid].buf1,UPCRI_IO_PAD), 
-                            addr, fetchsz);
+          _upcr_memget_nb((void *)UPCRI_ALIGNUP(lmbox_ref[threadid].buf1,UPCRI_IO_PAD), 
+                            addr, fetchsz UPCRI_PT_PASS);
         break;
       }
       case UPCRI_IO_NOENTRY: break;
@@ -1239,7 +1239,7 @@ static void upcri_process_mbox(upcr_pshared_ptr_t fd, int islocal, int isread UP
           size_t memvec_cnt;
           upc_shared_memvec_t *memvec;
 
-          bupc_waitsync(lmbox_ref[threadid].memcpy_handle);
+          _upcr_waitsync(lmbox_ref[threadid].memcpy_handle UPCRI_PT_PASS);
           lmbox_ref[threadid].memcpy_handle = BUPC_COMPLETE_HANDLE;
 
           upcri_metadata_mbox_extractvecs((upcri_metadata_mbox_t *)UPCRI_ALIGNUP(lmbox_ref[threadid].buf1,UPCRI_IO_PAD),
@@ -1275,12 +1275,12 @@ static void upcri_process_mbox(upcr_pshared_ptr_t fd, int islocal, int isread UP
   for (threadid = 0; threadid < upcr_threads(); threadid++) {
     if (lmbox_ref[threadid].type != UPCRI_IO_NOENTRY) {
       /* complete any pending data fetches */
-      bupc_waitsync(lmbox_ref[threadid].memcpy_handle);
+      _upcr_waitsync(lmbox_ref[threadid].memcpy_handle UPCRI_PT_PASS);
       lmbox_ref[threadid].memcpy_handle = BUPC_COMPLETE_HANDLE;
       if (lmbox_ref[threadid].memcpy_handle2 != NULL) {
         int hidx;
         for (hidx = 0; hidx < lmbox_ref[threadid].memcpy_handle2_cnt; hidx++) {
-          bupc_waitsync(lmbox_ref[threadid].memcpy_handle2[hidx]);
+          _upcr_waitsync(lmbox_ref[threadid].memcpy_handle2[hidx] UPCRI_PT_PASS);
         }
         if (lmbox_ref[threadid].type == UPCRI_IO_SINGLEENTRY) /* METADATA only, this is not a malloc buf */
           upcri_free(lmbox_ref[threadid].memcpy_handle2);
@@ -1412,7 +1412,7 @@ static void upcri_process_mbox(upcr_pshared_ptr_t fd, int islocal, int isread UP
             if (prev_mboxref->type != UPCRI_IO_NOENTRY && prev_mboxref->memcpy_handle3) {
               int hidx;
               for (hidx = 0; hidx < prev_mboxref->memcpy_handle3_cnt; hidx++) {
-                bupc_waitsync(prev_mboxref->memcpy_handle3[hidx]);
+                _upcr_waitsync(prev_mboxref->memcpy_handle3[hidx] UPCRI_PT_PASS);
               }
               upcri_free(prev_mboxref->memcpy_handle3);
               prev_mboxref->memcpy_handle3 = NULL;
@@ -1425,22 +1425,22 @@ static void upcri_process_mbox(upcr_pshared_ptr_t fd, int islocal, int isread UP
         }
         if (lmbox_ref[threadid].errno_result_mboxref != 0) { /* hit an I/O error - report to initiator in fd->errno_result */
           lmbox_ref[threadid].memcpy_handle3[lmbox_ref[threadid].memcpy_handle3_cnt] = 
-            bupc_memput_async(upcr_pshared_to_shared(
+            _upcr_memput_nb(upcr_pshared_to_shared(
                                upcr_add_psharedI(upcr_add_pshared1(fd, sizeof(bupc_file_t), threadid),
                                                   1, offsetof(bupc_file_t, errno_result_perthread))),
                                 &(lmbox_ref[threadid].errno_result_mboxref),
-                                sizeof(int));
+                                sizeof(int) UPCRI_PT_PASS);
           lmbox_ref[threadid].memcpy_handle3_cnt++;
           upcri_assert(lmbox_ref[threadid].memcpy_handle3_cnt <= handle_sz);
         } else if (lmbox_ref[threadid].size_result != lmbox_ref[threadid].data_len) {
           /* read hit EOF, write back modified result to reader's fd->size_result */
           upcri_assert(isread && lmbox_ref[threadid].size_result < lmbox_ref[threadid].data_len);
           lmbox_ref[threadid].memcpy_handle3[lmbox_ref[threadid].memcpy_handle3_cnt] = 
-            bupc_memput_async(upcr_pshared_to_shared(
+            _upcr_memput_nb(upcr_pshared_to_shared(
                                upcr_add_psharedI(upcr_add_pshared1(fd, sizeof(bupc_file_t), threadid),
                                                 1, offsetof(bupc_file_t, size_result))),
                               &(lmbox_ref[threadid].size_result),
-                              sizeof(lmbox_ref[threadid].size_result));
+                              sizeof(lmbox_ref[threadid].size_result) UPCRI_PT_PASS);
           lmbox_ref[threadid].memcpy_handle3_cnt++;
           upcri_assert(lmbox_ref[threadid].memcpy_handle3_cnt <= handle_sz);
         }
@@ -1453,7 +1453,7 @@ static void upcri_process_mbox(upcr_pshared_ptr_t fd, int islocal, int isread UP
       if (lmbox_ref[threadid].memcpy_handle3 != NULL) {
         int hidx;
         for (hidx = 0; hidx < lmbox_ref[threadid].memcpy_handle3_cnt; hidx++) {
-          bupc_waitsync(lmbox_ref[threadid].memcpy_handle3[hidx]);
+          _upcr_waitsync(lmbox_ref[threadid].memcpy_handle3[hidx] UPCRI_PT_PASS);
         }
         upcri_free(lmbox_ref[threadid].memcpy_handle3);
       }
