@@ -23,6 +23,15 @@
 GASNETT_INLINE(_upcri_newhandle)
 bupc_handle_t _upcri_newhandle(size_t extraspace UPCRI_PT_ARG) {
   bupc_handle_t h;
+#if UPCRI_SUPPORT_PTHREADS && !UPCRI_UPC_PTHREADS
+  /* Potentially have multiple phtreads per UPC thread.
+     TODO: rather than disable the freelist we could either:
+           + mutex-protect the one freelist
+           + implement per-pthread freelists
+   */
+  h = upcri_malloc(sizeof(upcri_eop)+extraspace);
+  h->next = BUPC_COMPLETE_HANDLE;
+#else
   if (extraspace == 0 && (h = upcri_auxdata()->eop_freelist) != NULL) {
     upcri_auxdata()->eop_freelist = h->next;
     h->next = UPCRI_TINY_EOP;
@@ -30,6 +39,7 @@ bupc_handle_t _upcri_newhandle(size_t extraspace UPCRI_PT_ARG) {
     h = upcri_malloc(sizeof(upcri_eop)+extraspace);
     h->next = extraspace ? BUPC_COMPLETE_HANDLE : UPCRI_TINY_EOP;
   }
+#endif
   return h;
 }
 
@@ -38,18 +48,23 @@ GASNETT_INLINE(_upcri_freehandlelist)
 void _upcri_freehandlelist(bupc_handle_t h UPCRI_PT_ARG) {
   while (h != BUPC_COMPLETE_HANDLE) {
     bupc_handle_t const next = h->next;
+#if UPCRI_SUPPORT_PTHREADS && !UPCRI_UPC_PTHREADS
+    upcri_assert(next != UPCRI_TINY_EOP);
+#else
     if (next == UPCRI_TINY_EOP) { /* end of list is a tiny eop */
       h->next = upcri_auxdata()->eop_freelist;
       upcri_auxdata()->eop_freelist = h;
       break;
-    } else {
+    } else
+#endif
+    {
       upcri_free(h);
     }
     h = next;
   }
 }
 
-extern void _bupc_waitsync(bupc_handle_t h UPCRI_PT_ARG) {
+extern void _upcr_waitsync(bupc_handle_t h UPCRI_PT_ARG) {
   bupc_handle_t hp;
   if (h == BUPC_COMPLETE_HANDLE) return;
   upcri_assert(h != UPCRI_TINY_EOP);
@@ -60,7 +75,7 @@ extern void _bupc_waitsync(bupc_handle_t h UPCRI_PT_ARG) {
   }
   upcri_freehandlelist(h);
 }
-extern int _bupc_trysync(bupc_handle_t h UPCRI_PT_ARG) {
+extern int _upcr_trysync(bupc_handle_t h UPCRI_PT_ARG) {
   bupc_handle_t hp;
   if (h == BUPC_COMPLETE_HANDLE) return 1;
   upcri_assert(h != UPCRI_TINY_EOP);
@@ -79,7 +94,7 @@ extern void _bupc_waitsync_all (bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG) {
   size_t i;
   upcri_assert(_n > 0 && _ph);
   for (i=0; i < _n; i++) {
-    _bupc_waitsync(_ph[i] UPCRI_PT_PASS);
+    _upcr_waitsync(_ph[i] UPCRI_PT_PASS);
     _ph[i] = BUPC_COMPLETE_HANDLE;
   }
 }
@@ -89,7 +104,7 @@ extern int  _bupc_trysync_all  (bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG) {
   for (i=0; i < _n; i++) {
     bupc_handle_t const h = _ph[i];
     if (h == BUPC_COMPLETE_HANDLE) ;
-    else if (_bupc_trysync(h UPCRI_PT_PASS)) { /* found complete */
+    else if (_upcr_trysync(h UPCRI_PT_PASS)) { /* found complete */
       _ph[i] = BUPC_COMPLETE_HANDLE;
     } else { /* found incomplete */
       return 0;
@@ -113,7 +128,7 @@ extern int  _bupc_trysync_some (bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG) {
     bupc_handle_t const h = _ph[i];
     if (h != BUPC_COMPLETE_HANDLE) {
       empty = 0;
-      if (_bupc_trysync(h UPCRI_PT_PASS)) { /* found complete */
+      if (_upcr_trysync(h UPCRI_PT_PASS)) { /* found complete */
         _ph[i] = BUPC_COMPLETE_HANDLE;
         success = 1;
       }
@@ -136,19 +151,19 @@ extern int  _bupc_trysync_some (bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG) {
     }                                                              \
   } while (0)
 
-extern bupc_handle_t _bupc_memcpy_async(upcr_shared_ptr_t dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) {
+extern bupc_handle_t _upcr_memcpy_nb(upcr_shared_ptr_t dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) {
   UPCRI_EOP_ENCAPSULATE(_upcr_nb_memcpy(dst, src, n UPCRI_PT_PASS));
 }
 
-extern bupc_handle_t _bupc_memget_async(void *dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) {
+extern bupc_handle_t _upcr_memget_nb(void *dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) {
   UPCRI_EOP_ENCAPSULATE(_upcr_nb_memget(dst, src, n UPCRI_PT_PASS));
 }
 
-extern bupc_handle_t _bupc_memput_async(upcr_shared_ptr_t dst, const void *src, size_t n UPCRI_PT_ARG) {
+extern bupc_handle_t _upcr_memput_nb(upcr_shared_ptr_t dst, const void *src, size_t n UPCRI_PT_ARG) {
   UPCRI_EOP_ENCAPSULATE(_upcr_nb_memput(dst, (void *)src, n UPCRI_PT_PASS));
 }
 
-extern bupc_handle_t _bupc_memset_async(upcr_shared_ptr_t dst, int c, size_t n UPCRI_PT_ARG) {
+extern bupc_handle_t _upcr_memset_nb(upcr_shared_ptr_t dst, int c, size_t n UPCRI_PT_ARG) {
   UPCRI_EOP_ENCAPSULATE(_upcr_nb_memset(dst, c, n UPCRI_PT_PASS));
 }
 
@@ -282,7 +297,7 @@ typedef struct {
       char *tmpdata = upcri_eop_metadata(h)+sizeof(bupc_pmemvec_t);
       tmplist->addr = tmpdata;
       tmplist->len = datasz;
-      _bupc_waitsync(
+      _upcr_waitsync(
         _bupc_memget_vlist_async(1, tmplist, srccount, srclist 
         UPCRI_PT_PASS) UPCRI_PT_PASS);
       h->handle = GASNET_INVALID_HANDLE; /* this entry just holds the bounce buffer and metadata memory */
@@ -538,7 +553,7 @@ typedef struct {
       void **tmplist = (void **)upcri_eop_metadata(h);
       char *tmpdata = upcri_eop_metadata(h)+sizeof(void *);
       *tmplist = tmpdata;
-      _bupc_waitsync(
+      _upcr_waitsync(
         _bupc_memget_ilist_async(1, tmplist, datasz, srccount, srclist, srclen 
         UPCRI_PT_PASS) UPCRI_PT_PASS);
       h->handle = GASNET_INVALID_HANDLE; /* this entry just holds the bounce buffer and metadata memory */
@@ -916,7 +931,7 @@ typedef struct {
         size_t *count = (size_t*)(upcri_eop_metadata(h)+2*sizeof(size_t));
         char *tmpdata = upcri_eop_metadata(h)+4*sizeof(size_t);
         /* blocking get */
-        _bupc_waitsync(
+        _upcr_waitsync(
           _bupc_memget_fstrided_async(tmpdata, datasz, datasz, 1,
                                      srcaddr, srcchunklen, srcchunkstride, srcchunkcount
                                      UPCRI_PT_PASS) UPCRI_PT_PASS);
@@ -1086,7 +1101,7 @@ typedef struct {
         size_t strsz = 1;
         for (i = 0; i < stridelevels; i++) { strsz *= count[i]; strides[i] = strsz; }
         /* blocking get */
-        _bupc_waitsync(
+        _upcr_waitsync(
           _bupc_memget_strided_async(tmpdata, strides, 
                                      srcaddr, srcstrides, 
                                      count, stridelevels
