@@ -42,33 +42,71 @@ typedef struct _upcri_eop *bupc_handle_t;
 #endif
 
 /*---------------------------------------------------------------------------------*/
-extern void _bupc_waitsync(bupc_handle_t h UPCRI_PT_ARG);
-extern int _bupc_trysync(bupc_handle_t h UPCRI_PT_ARG);
+
+#define upcr_waitsync(h) \
+       (upcri_srcpos(), _upcr_waitsync(h UPCRI_PT_PASS))
+extern void _upcr_waitsync(bupc_handle_t h UPCRI_PT_ARG);
+#define upcr_trysync(h) \
+       (upcri_srcpos(), _upcr_trysync(h UPCRI_PT_PASS))
+extern int _upcr_trysync(bupc_handle_t h UPCRI_PT_ARG);
+
+#define upcr_memcpy_nb(dst,src,n) \
+       (upcri_srcpos(), _upcr_memcpy_nb(dst,src,n UPCRI_PT_PASS))
+extern bupc_handle_t _upcr_memcpy_nb(upcr_shared_ptr_t dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
+#define upcr_memget_nb(dst,src,n) \
+       (upcri_srcpos(), _upcr_memget_nb(dst,src,n UPCRI_PT_PASS))
+extern bupc_handle_t _upcr_memget_nb(void *dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
+#define upcr_memput_nb(dst,src,n) \
+       (upcri_srcpos(), _upcr_memput_nb(dst,src,n UPCRI_PT_PASS))
+extern bupc_handle_t _upcr_memput_nb(upcr_shared_ptr_t dst, const void *src, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
+#define upcr_memset_nb(dst,c,n) \
+       (upcri_srcpos(), _upcr_memset_nb(dst,c,n UPCRI_PT_PASS))
+extern bupc_handle_t _upcr_memset_nb(upcr_shared_ptr_t dst, int c, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
+
 extern void _bupc_waitsync_all (bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG);
 extern int  _bupc_trysync_all  (bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG);
 extern void _bupc_waitsync_some(bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG);
 extern int  _bupc_trysync_some (bupc_handle_t *_ph, size_t _n UPCRI_PT_ARG);
 
-extern bupc_handle_t _bupc_memcpy_async(upcr_shared_ptr_t dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
-extern bupc_handle_t _bupc_memget_async(void *dst, upcr_shared_ptr_t src, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
-extern bupc_handle_t _bupc_memput_async(upcr_shared_ptr_t dst, const void *src, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
-extern bupc_handle_t _bupc_memset_async(upcr_shared_ptr_t dst, int c, size_t n UPCRI_PT_ARG) GASNETT_WARN_UNUSED_RESULT;
-
-#define _bupc_waitsynci() upcr_wait_syncnbi_all()
-#define _bupc_trysynci()  upcr_try_syncnbi_all()
-#define _bupc_memget_asynci   upcr_nbi_memget
-#define _bupc_memput_asynci   upcr_nbi_memput
-#define _bupc_memcpy_asynci   upcr_nbi_memcpy
-#define _bupc_memset_asynci   upcr_nbi_memset
-
 #define _bupc_begin_accessregion() upcr_begin_nbi_accessregion()
 extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
 
-#if !UPCRI_LIBWRAP 
-  #define bupc_waitsync(h) \
-       (upcri_srcpos(), _bupc_waitsync(h UPCRI_PT_PASS))
-  #define bupc_trysync(h) \
-       (upcri_srcpos(), _bupc_trysync(h UPCRI_PT_PASS))
+#if !UPCRI_LIBWRAP
+  UPCRI_DEPRECATED_STUB(bupc_memcpy_async)
+  #define bupc_memcpy_async(dst,src,n) \
+       (bupc_memcpy_async (), upcr_memcpy_nb(dst,src,n))
+  UPCRI_DEPRECATED_STUB(bupc_memget_async)
+  #define bupc_memget_async(dst,src,n) \
+       (bupc_memget_async (), upcr_memget_nb(dst,src,n))
+  UPCRI_DEPRECATED_STUB(bupc_memput_async)
+  #define bupc_memput_async(dst,src,n) \
+       (bupc_memput_async (), upcr_memput_nb(dst,src,n))
+  UPCRI_DEPRECATED_STUB(bupc_memset_async)
+  #define bupc_memset_async(dst,c,n) \
+       (bupc_memset_async (), upcr_memset_nb(dst,c,n))
+
+  UPCRI_DEPRECATED_STUB(bupc_waitsync)
+  #define bupc_waitsync(h) (bupc_waitsync (), upcr_waitsync(h))
+  UPCRI_DEPRECATED_STUB(bupc_trysync)
+  #define bupc_trysync(h)   (bupc_trysync (), upcr_trysync(h))
+
+  UPCRI_DEPRECATED_STUB(bupc_memcpy_asynci)
+  #define bupc_memcpy_asynci(dst,src,n) \
+       (bupc_memcpy_asynci (), upcr_nbi_memcpy(dst,src,n))
+  UPCRI_DEPRECATED_STUB(bupc_memget_asynci)
+  #define bupc_memget_asynci(dst,src,n) \
+       (bupc_memget_asynci (), upcr_nbi_memget(dst,src,n))
+  UPCRI_DEPRECATED_STUB(bupc_memput_asynci)
+  #define bupc_memput_asynci(dst,src,n) \
+       (bupc_memput_asynci (), upcr_nbi_memput(dst,src,n))
+  UPCRI_DEPRECATED_STUB(bupc_memset_asynci)
+  #define bupc_memset_asynci(dst,c,n) \
+       (bupc_memset_asynci (), upcr_nbi_memset(dst,c,n))
+
+  UPCRI_DEPRECATED_STUB(bupc_waitsynci)
+  #define bupc_waitsynci() (bupc_waitsynci (), upcr_wait_syncnbi_all())
+  UPCRI_DEPRECATED_STUB(bupc_trysynci)
+  #define bupc_trysynci()  (bupc_trysynci (), upcr_try_syncnbi_all())
 
   #define bupc_waitsync_all(ph,n) \
        (upcri_srcpos(), _bupc_waitsync_all((ph),(n) UPCRI_PT_PASS))
@@ -79,55 +117,15 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   #define bupc_trysync_some(ph,n) \
        (upcri_srcpos(), _bupc_trysync_some((ph),(n) UPCRI_PT_PASS))
 
-  #define bupc_memcpy_async(dst,src,n) \
-       (upcri_srcpos(), _bupc_memcpy_async(dst,src,n UPCRI_PT_PASS))
-  #define bupc_memget_async(dst,src,n) \
-       (upcri_srcpos(), _bupc_memget_async(dst,src,n UPCRI_PT_PASS))
-  #define bupc_memput_async(dst,src,n) \
-       (upcri_srcpos(), _bupc_memput_async(dst,src,n UPCRI_PT_PASS))
-  #define bupc_memset_async(dst,c,n) \
-       (upcri_srcpos(), _bupc_memset_async(dst,c,n UPCRI_PT_PASS))
-
-  #define bupc_waitsynci     _bupc_waitsynci
-  #define bupc_trysynci      _bupc_trysynci
-  #define bupc_memget_asynci _bupc_memget_asynci
-  #define bupc_memput_asynci _bupc_memput_asynci
-  #define bupc_memcpy_asynci _bupc_memcpy_asynci
-  #define bupc_memset_asynci _bupc_memset_asynci
-
   #define bupc_begin_accessregion _bupc_begin_accessregion
   #define bupc_end_accessregion() \
        (upcri_srcpos(), _bupc_end_accessregion(UPCRI_PT_PASS_ALONE))
-
-#elif UPCRI_BUILDING_LIBUPCR
-  #define bupc_waitsync     _bupc_waitsync
-  #define bupc_trysync      _bupc_trysync
-
-  #define bupc_waitsync_all  _bupc_waitsync_all
-  #define bupc_trysync_all   _bupc_trysync_all
-  #define bupc_waitsync_some _bupc_waitsync_some
-  #define bupc_trysync_some  _bupc_trysync_some
-
-  #define bupc_memcpy_async _bupc_memcpy_async
-  #define bupc_memget_async _bupc_memget_async
-  #define bupc_memput_async _bupc_memput_async
-  #define bupc_memset_async _bupc_memset_async
-
-  #define bupc_waitsynci     _bupc_waitsynci
-  #define bupc_trysynci      _bupc_trysynci
-  #define bupc_memget_asynci _bupc_memget_asynci
-  #define bupc_memput_asynci _bupc_memput_asynci
-  #define bupc_memcpy_asynci _bupc_memcpy_asynci
-  #define bupc_memset_asynci _bupc_memset_asynci
-
-  #define bupc_begin_accessregion _bupc_begin_accessregion
-  #define bupc_end_accessregion   _bupc_end_accessregion
-
 #endif
+
 /*---------------------------------------------------------------------------------*/
 
 #if BUPC_SG_DESIGN_A
-  #if UPCRI_LIBWRAP && __GCC_UPC__ /* UPC lang compiles */
+  #if UPCRI_LIBWRAP && (__GCC_UPC__ || __clang_upc__) /* UPC lang compiles */
   typedef struct bupc_smemvec_S_trans {
     shared void *addr;
     size_t len;
@@ -141,7 +139,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   typedef gasnet_memvec_t bupc_pmemvec_t;
   #endif
 
-  #if UPCRI_LIBWRAP && __GCC_UPC__ /* UPC lang compiles */
+  #if UPCRI_LIBWRAP && (__GCC_UPC__ || __clang_upc__) /* UPC lang compiles */
     #define UPCR_BLESS_SVP(a)  upcri_bless_SVP2shared(a)
     #define UPCR_BLESS_SVPI(a) (*(upcr_shared_ptr_t*)&(a))
     #define UPCR_BLESS_SVPA(a) upcri_bless_SVPA2sharedA(a)
@@ -222,7 +220,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   GASNETT_INLINE(_bupc_memcpy_vlist)
   void _bupc_memcpy_vlist(size_t dstcount, bupc_smemvec_t const dstlist[], 
                         size_t srccount, bupc_smemvec_t const srclist[] UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memcpy_vlist_async(dstcount, dstlist, srccount, srclist UPCRI_PT_PASS) UPCRI_PT_PASS);
+    _upcr_waitsync(_bupc_memcpy_vlist_async(dstcount, dstlist, srccount, srclist UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memcpy_vlist(dstcount,dstlist,srccount,srclist)                       \
          (upcri_srcpos(), upcri_check_vlists(s,dstcount,dstlist,s,srccount,srclist), \
@@ -231,7 +229,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   GASNETT_INLINE(_bupc_memput_vlist)
   void _bupc_memput_vlist(size_t dstcount, bupc_smemvec_t const dstlist[], 
                        size_t srccount, bupc_pmemvec_t const srclist[] UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memput_vlist_async(dstcount, dstlist, srccount, srclist UPCRI_PT_PASS) UPCRI_PT_PASS);
+    _upcr_waitsync(_bupc_memput_vlist_async(dstcount, dstlist, srccount, srclist UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memput_vlist(dstcount,dstlist,srccount,srclist)                       \
          (upcri_srcpos(), upcri_check_vlists(s,dstcount,dstlist,p,srccount,srclist), \
@@ -240,7 +238,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   GASNETT_INLINE(_bupc_memget_vlist)
   void _bupc_memget_vlist(size_t dstcount, bupc_pmemvec_t const dstlist[], 
                        size_t srccount, bupc_smemvec_t const srclist[] UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memget_vlist_async(dstcount, dstlist, srccount, srclist UPCRI_PT_PASS) UPCRI_PT_PASS);
+    _upcr_waitsync(_bupc_memget_vlist_async(dstcount, dstlist, srccount, srclist UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memget_vlist(dstcount,dstlist,srccount,srclist)                       \
          (upcri_srcpos(), upcri_check_vlists(p,dstcount,dstlist,s,srccount,srclist), \
@@ -307,7 +305,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   GASNETT_INLINE(_bupc_memcpy_ilist)
   void _bupc_memcpy_ilist(size_t dstcount, upcr_shared_ptr_t const dstlist[], size_t dstlen,
                         size_t srccount, upcr_shared_ptr_t const srclist[], size_t srclen UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memcpy_ilist_async(dstcount, dstlist, dstlen, srccount, srclist, srclen UPCRI_PT_PASS) UPCRI_PT_PASS);
+    _upcr_waitsync(_bupc_memcpy_ilist_async(dstcount, dstlist, dstlen, srccount, srclist, srclen UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memcpy_ilist(dstcount,dstlist,dstlen,srccount,srclist,srclen)                       \
          (upcri_srcpos(), upcri_check_ilists(s,dstcount,UPCR_BLESS_SVPA(dstlist),dstlen,s,srccount,UPCR_BLESS_SVPA(srclist),srclen), \
@@ -316,7 +314,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   GASNETT_INLINE(_bupc_memput_ilist)
   void _bupc_memput_ilist(size_t dstcount, upcr_shared_ptr_t const dstlist[], size_t dstlen,
                         size_t srccount,        const void * const srclist[], size_t srclen UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memput_ilist_async(dstcount, dstlist, dstlen, srccount, srclist, srclen UPCRI_PT_PASS) UPCRI_PT_PASS);
+    _upcr_waitsync(_bupc_memput_ilist_async(dstcount, dstlist, dstlen, srccount, srclist, srclen UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memput_ilist(dstcount,dstlist,dstlen,srccount,srclist,srclen)                       \
          (upcri_srcpos(), upcri_check_ilists(s,dstcount,UPCR_BLESS_SVPA(dstlist),dstlen,p,srccount,srclist,srclen), \
@@ -325,7 +323,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   GASNETT_INLINE(_bupc_memget_ilist)
   void _bupc_memget_ilist(size_t dstcount,              void * const dstlist[], size_t dstlen, 
                         size_t srccount, upcr_shared_ptr_t const srclist[], size_t srclen UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memget_ilist_async(dstcount, dstlist, dstlen, srccount, srclist, srclen UPCRI_PT_PASS) UPCRI_PT_PASS);
+    _upcr_waitsync(_bupc_memget_ilist_async(dstcount, dstlist, dstlen, srccount, srclist, srclen UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memget_ilist(dstcount,dstlist,dstlen,srccount,srclist,srclen)                       \
          (upcri_srcpos(), upcri_check_ilists(p,dstcount,dstlist,dstlen,s,srccount,UPCR_BLESS_SVPA(srclist),srclen), \
@@ -398,7 +396,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
                            size_t dstchunkstride, size_t dstchunkcount,
                            upcr_shared_ptr_t srcaddr,  size_t srcchunklen, 
                            size_t srcchunkstride, size_t srcchunkcount UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memcpy_fstrided_async(dstaddr, dstchunklen, dstchunkstride, dstchunkcount,
+    _upcr_waitsync(_bupc_memcpy_fstrided_async(dstaddr, dstchunklen, dstchunkstride, dstchunkcount,
                                             srcaddr, srcchunklen, srcchunkstride, srcchunkcount UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memcpy_fstrided(dstaddr,dstchunklen,dstchunkstride,dstchunkcount,srcaddr,srcchunklen,srcchunkstride,srcchunkcount)                       \
@@ -410,7 +408,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
                            size_t dstchunkstride, size_t dstchunkcount,
                                   void *srcaddr,  size_t srcchunklen, 
                            size_t srcchunkstride, size_t srcchunkcount UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memput_fstrided_async(dstaddr, dstchunklen, dstchunkstride, dstchunkcount,
+    _upcr_waitsync(_bupc_memput_fstrided_async(dstaddr, dstchunklen, dstchunkstride, dstchunkcount,
                                             srcaddr, srcchunklen, srcchunkstride, srcchunkcount UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memput_fstrided(dstaddr,dstchunklen,dstchunkstride,dstchunkcount,srcaddr,srcchunklen,srcchunkstride,srcchunkcount)                       \
@@ -422,7 +420,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
                            size_t dstchunkstride, size_t dstchunkcount,
                            upcr_shared_ptr_t srcaddr,  size_t srcchunklen, 
                            size_t srcchunkstride, size_t srcchunkcount UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memget_fstrided_async(dstaddr, dstchunklen, dstchunkstride, dstchunkcount,
+    _upcr_waitsync(_bupc_memget_fstrided_async(dstaddr, dstchunklen, dstchunkstride, dstchunkcount,
                                             srcaddr, srcchunklen, srcchunkstride, srcchunkcount UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
   #define bupc_memget_fstrided(dstaddr,dstchunklen,dstchunkstride,dstchunkcount,srcaddr,srcchunklen,srcchunkstride,srcchunkcount)                       \
@@ -494,7 +492,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   void _bupc_memcpy_strided(upcr_shared_ptr_t dstaddr, const size_t dststrides[], 
                            upcr_shared_ptr_t srcaddr, const size_t srcstrides[], 
                            const size_t count[], size_t stridelevels UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memcpy_strided_async(dstaddr, dststrides,
+    _upcr_waitsync(_bupc_memcpy_strided_async(dstaddr, dststrides,
                                             srcaddr, srcstrides, 
                                             count, stridelevels UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
@@ -506,7 +504,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   void _bupc_memput_strided(upcr_shared_ptr_t dstaddr, const size_t dststrides[], 
                                   const void *srcaddr, const size_t srcstrides[], 
                            const size_t count[], size_t stridelevels UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memput_strided_async(dstaddr, dststrides,
+    _upcr_waitsync(_bupc_memput_strided_async(dstaddr, dststrides,
                                             srcaddr, srcstrides, 
                                             count, stridelevels UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
@@ -518,7 +516,7 @@ extern bupc_handle_t _bupc_end_accessregion(UPCRI_PT_ARG_ALONE);
   void _bupc_memget_strided(             void *dstaddr, const size_t dststrides[], 
                            upcr_shared_ptr_t srcaddr, const size_t srcstrides[], 
                            const size_t count[], size_t stridelevels UPCRI_PT_ARG) {
-    _bupc_waitsync(_bupc_memget_strided_async(dstaddr, dststrides,
+    _upcr_waitsync(_bupc_memget_strided_async(dstaddr, dststrides,
                                             srcaddr, srcstrides, 
                                             count, stridelevels UPCRI_PT_PASS) UPCRI_PT_PASS);
   }
