@@ -40,18 +40,6 @@
 
 #include <upcr_preinclude/upc_bits.h>
 
-/* The following pre-processor definitions are required by section 7.1
-   in the UPC 1.0 specification.  They were removed in UPC Spec 1.1,
-   but we'll optionally keep them for backwards compatibility
-*/
-#if UPCR_ALLOW_UPC_1_0_KEYWORDS
-#define barrier upc_barrier
-#define barrier_notify upc_notify
-#define barrier_wait upc_wait
-#define forall upc_forall
-#define fence upc_fence
-#endif
-
 #if ! UPC_MAX_BLOCK_SIZE
 #  error UPC_MAX_BLOCK_SIZE is not properly defined
 #endif
@@ -87,7 +75,6 @@ extern size_t upc_addrfield (shared void *);
 
 extern shared void *upc_global_alloc (size_t, size_t);
 extern shared void *upc_all_alloc (size_t, size_t);
-extern shared void *upc_local_alloc (size_t, size_t); /* XXX: how long to keep this - removed in UPC 1.3? */
 extern shared void *upc_alloc(size_t);
 extern void upc_free (shared void *);
 extern void upc_all_free (shared void *);
