@@ -25,6 +25,9 @@ uintptr_t	upcri_linksegstart;
   #if !UPCRI_UPC_PTHREADS
   uintptr_t	upcri_linkoffset_single;
   #endif
+  #define UPCRI_LINKSEGSTART upcri_linksegstart
+#else
+  #define UPCRI_LINKSEGSTART 0
 #endif
 
 #if GASNET_TRACE
@@ -41,14 +44,19 @@ _upcr_do_memcpy(upcr_shared_ptr_t dst, upcr_shared_ptr_t src,
 {
     UPCRI_PASS_GAS();
 
-#if (GASNET_CONDUIT_GEMINI || GASNET_CONDUIT_ARIES) && GASNET_PSHM
-    /* Work around bug 3178 w.r.t. dynamic registration of other-node's segment */
-    const gasnet_node_t thisnode = gasnet_mynode();
-    upcri_local_t srclocal = (upcri_shared_nodeof(src) == thisnode) ? upcri_s_islocal(src) : 0;
-    upcri_local_t dstlocal = (upcri_shared_nodeof(dst) == thisnode) ? upcri_s_islocal(dst) : 0;
-#else
     upcri_local_t srclocal = upcri_s_islocal(src);
     upcri_local_t dstlocal = upcri_s_islocal(dst);
+
+#if (GASNET_CONDUIT_GEMINI || GASNET_CONDUIT_ARIES) && GASNET_PSHM
+    /* Work around bug 3178 w.r.t. dynamic registration of other-node's segment.
+     * While that bug has been fixed (as far as correctness goes), there is still a
+     * performance hit from failed registration that we can avoid by checking here.
+     */
+    if (!srclocal || !dstlocal) { /* Some communication required */
+        const gasnet_node_t thisnode = gasnet_mynode();
+        if (upcri_shared_nodeof(src) != thisnode) srclocal = 0;
+        if (upcri_shared_nodeof(dst) != thisnode) dstlocal = 0;
+    }
 #endif
 
     if (srclocal && dstlocal) { /* no communication */
@@ -62,6 +70,7 @@ _upcr_do_memcpy(upcr_shared_ptr_t dst, upcr_shared_ptr_t src,
 	    case upcri_memcpy_type_nb:
 		return gasnet_put_nb_bulk(dnode, daddr, saddr, nbytes);
 	    case upcri_memcpy_type_block:
+		/* TODO: implement UPCRI_DO_MEMPUT_NBI case */
 		gasnet_put_bulk(dnode, daddr, saddr, nbytes);
 		break;
 	    case upcri_memcpy_type_nbi:
@@ -85,6 +94,7 @@ _upcr_do_memcpy(upcr_shared_ptr_t dst, upcr_shared_ptr_t src,
     } else { /* fully remote */
       gasnet_node_t snode = upcri_shared_nodeof(src);
       gasnet_node_t dnode = upcri_shared_nodeof(dst);
+      /* TODO: "co-located" case could be reimplemented at super-node level */
       if (snode == dnode) { /* remotely co-located, use AM */
 	  return upcri_do_remote_memcpy(snode, upcri_shared_to_remote(dst), upcri_shared_to_remote(src), nbytes);
       } else { /* separately remote */
@@ -96,6 +106,10 @@ _upcr_do_memcpy(upcr_shared_ptr_t dst, upcr_shared_ptr_t src,
 	 * truly nonblocking.  But we don't care much about the performance of
 	 * 3rd party memcpy's anyway, as they should seldom be used by
 	 * sensible programmers.
+	 *
+	 * TODO: GASNet-EX progress functions will allow an AM-based
+	 * implementation in which the src (dst) node uses a progress
+	 * function queued in the AM handler to issue a Put (Get).
 	 */
     	upcr_shared_ptr_t buf = upcr_alloc(nbytes);
 	void *pbuf = upcr_shared_to_local(buf);
@@ -139,9 +153,14 @@ upcr_shared_ptr_t _bupc_local_to_shared(void *ptr, int thread, int phase)
  * buffer.  Returns the address of the buffer, or of an error string if the
  * buffer was too short.  The buffer passed should be at least
  * UPCRI_DUMP_MIN_LENGTH in length.
+ *
+ * AM-SAFE
  */
 int _bupc_dump_shared(upcr_shared_ptr_t ptr, char *buf, int maxlen) {
+#if 0
+    /* This is not needed, and would preclude AM-SAFETY */
     UPCR_BEGIN_FUNCTION();
+#endif
 
     if (maxlen < UPCRI_DUMP_MIN_LENGTH) {
 	if (maxlen > 15)
@@ -382,42 +401,41 @@ int _upcri_isvalid_pshared(upcr_pshared_ptr_t sptr) {
  */
 upcr_shared_ptr_t _bupc_inverse_cast(void *ptr)
 {
-    const uintptr_t addr = (uintptr_t)ptr;
     upcr_shared_ptr_t s = upcr_null_shared;
+    uintptr_t addrfield = 0;
+    int thread = -1;
 
     if_pt (ptr != NULL) {
       #if UPCRI_SHARED_THREADS
         /* binary search addressable shared heaps for a candidate */
+        const uintptr_t seg_addr = (uintptr_t)ptr - UPCRI_LINKSEGSTART;
         int lo = 0;
         int hi = upcri_shared_threads;
         int mid;
         while (lo != (mid = (lo + hi) >> 1)) {
             const int th = upcri_shared_thread[mid];
-            if (upcri_thread2local[th] <= addr) {
+            if (upcri_thread2local[th] <= seg_addr) {
                 lo = mid;
             } else {
                 hi = mid;
             }
         }
-        /* bounds check against the candidate */
-        {
-            const int th = upcri_shared_thread[mid];
-            const uintptr_t offset = addr - upcri_thread2local[th];
-            if_pf (offset >= UPCR_PAGESIZE && offset < upcri_perthread_segsize) {
-                s = upcri_addrfield_to_shared(offset, th, 0);
-            }
-        }
+        thread = upcri_shared_thread[mid];
+        addrfield = (uintptr_t)ptr - upcri_thread2local[thread];
         upcri_assert(UPCRI_SPTRS_USE_OFFSETS); /* or above computation is incorrect */
       #else
         /* only my own shared heap is addressable - just check it */
-        const uintptr_t offset = addr - upcri_myregion_single;
-        if_pf (offset >= UPCR_PAGESIZE && offset < upcri_perthread_segsize) {
-            s = upcri_addrfield_to_shared(addr UPCRI_MINUS_MY_OFFSET, upcr_mythread(), 0);
-        }
+        thread = upcr_mythread();
+        addrfield = (uintptr_t)ptr UPCRI_MINUS_MY_OFFSET;
       #endif
+
+      /* bounds check against the candidate */
+      if_pf (_upcri_check_addr_bounds(addrfield, thread)) {
+          s = upcri_addrfield_to_shared(addrfield, thread, 0);
+      }
     }
 
-    upcri_assert(upcr_isnull_shared(s) || (ptr == _bupc_cast(s)));
+    upcri_assert(upcr_isnull_shared(s) || (ptr == upcr_cast(s)));
     return s;
 }
 
