@@ -63,6 +63,15 @@ GASNETT_BEGIN_EXTERNC
 #	error UPCRI_SINGLE_ALIGNED_REGIONS must be defined to 0 or 1!
 #endif
 
+
+/*
+ * At least GASNet's udp-conduit includes sockets.h and at least Solaris 11.2
+ * #defines s_addr there, causing big problems when we use it a field name.
+ */
+#ifdef s_addr
+  #undef s_addr
+#endif
+
 /* 
  * "Public size" for phase:  shared pointers may store something smaller.
  */
@@ -78,7 +87,7 @@ typedef unsigned int upcr_phase_t;
  */
 
 #if defined(UPCRI_SYMMETRIC_SPTR) && !defined(UPCRI_SYMMETRIC_PSHARED)
-  #if UPCRI_SYMMETRIC_SEGMENTS && !UPCRI_USING_GCCUPC
+  #if UPCRI_SYMMETRIC_SEGMENTS && !(UPCRI_USING_GCCUPC || UPCRI_USING_CUPC)
     #define UPCRI_SYMMETRIC_PSHARED 1
   #else
     #define UPCRI_SYMMETRIC_PSHARED 0
@@ -172,7 +181,7 @@ typedef unsigned int upcr_phase_t;
   #if (UPCRI_PHASE_BITS + UPCRI_THREAD_BITS + UPCRI_ADDR_BITS) != 64
     #error Packed shared pointer components do not add up to 64 bits!
   #endif
-  #if UPCRI_USING_GCCUPC
+  #if UPCRI_USING_GCCUPC || UPCRI_USING_CUPC
     /* Presently GCCUPC uses 36 bits for addr, even on 32-bit arch */
   #elif UPCRI_ADDR_BITS > SIZEOF_VOID_P*8
     #error UPCRI_ADDR_BITS is larger than 8*sizeof(void*)
@@ -186,7 +195,7 @@ typedef unsigned int upcr_phase_t;
   #if UPCRI_ADDR_BITS < 20
     #error UPCRI_ADDR_BITS is less than 20 bits
   #endif
-  #if UPCRI_USING_GCCUPC && UPCRI_SPTR_ADDR_FIRST
+  #if (UPCRI_USING_GCCUPC || UPCRI_USING_CUPC) && UPCRI_SPTR_ADDR_FIRST
     /* GCCUPC packed pointer: high                 low
                               | addr | thread | phase |
        Different representation then BUPC is chosen because it
@@ -223,7 +232,7 @@ typedef unsigned int upcr_phase_t;
     #define upcri_check_addr_overflow(addrfield)
   #endif
 
-  #if UPCRI_USING_GCCUPC
+  #if UPCRI_USING_GCCUPC || UPCRI_USING_CUPC
     /* GCCUPC puts offsets only in addr field */
     #define UPCRI_ADDR_FITS_IN_SPTR   0
   #else
@@ -232,7 +241,7 @@ typedef unsigned int upcr_phase_t;
 
 #elif UPCRI_STRUCT_SPTR
 
-  #if UPCRI_USING_GCCUPC
+  #if UPCRI_USING_GCCUPC || UPCRI_USING_CUPC || UPCRI_USING_CUPC2C
 
     /* Make sure that configure set the appropiate
        field bit size values.  */
@@ -268,7 +277,7 @@ typedef unsigned int upcr_phase_t;
     #endif
 
     typedef struct
-#if __GCC_UPC__
+#if __GCC_UPC__ || __clang_upc__
     /* Make sure that UPCR's shared pointer struct type isn't
        allowed to 'pun' either GCCUPC's internal shared pointer representation
        type, or pointer-to-shared types.  (See Bug #2424 for details.) */
@@ -300,9 +309,10 @@ typedef unsigned int upcr_phase_t;
       __attribute__ ((__aligned__ (UPCRI_STRUCT_SPTR_ALIGN)))
 #endif
       upcr_shared_ptr_t;
-    /* pshared ptrs currently not used by GCCUPC */
+    /* pshared ptrs currently not used by GNU UPC or Clang UPC */
     typedef struct shared_ptr_struct upcr_pshared_ptr_t;
 
+    /* TODO: should these adjust for (!UPCRI_SPTR_ADDR_FIRST) ? */
     #define UPCR_NULL_SHARED	    { 0, 0, 0 }
     #define UPCR_NULL_PSHARED	    { 0, 0 }
     #define UPCR_INITIALIZED_SHARED   { 0, 0, 1 }
@@ -343,7 +353,7 @@ typedef unsigned int upcr_phase_t;
     #endif
 
     #define UPCRI_ADDR_FITS_IN_SPTR   1
-  #endif /* UPCRI_USING_GCCUPC */
+  #endif /* UPCRI_USING_GCCUPC || UPCRI_USING_CUPC */
 
 #else
   #error shared pointer representation not defined!
