@@ -47,7 +47,7 @@ my %conf = (
     enable_fca_by_default   => '',
     enable_fca              => 0
 );
-my (@all_conduits) = ('mpi', 'udp', 'smp', 'ibv', 'shmem', 'portals4', 'gemini', 'aries', 'pami', 'mxm');
+my (@all_conduits) = ('mpi', 'udp', 'smp', 'ibv', 'shmem', 'portals4', 'gemini', 'aries', 'pami', 'ofi', 'mxm');
 foreach (@all_conduits) {
     $conf{$_."_spawn"}       = 'nodefault';
     $conf{$_."_spawn_nodes"} = '';
@@ -56,7 +56,7 @@ foreach (@all_conduits) {
 }
 
 # directories
-my ($cwd, $in_build_tree, $upcr_home, $upcr_bin, $upcr_etc, $upcr_lib, $upcr_include, $upcr_share);
+my ($cwd, $in_build_tree, $upcr_home, $upcr_bin, $upcr_etc, $upcr_include, $upcr_share);
 
 # Command-line switches
 my ( 
@@ -156,7 +156,7 @@ sub getOpts
 	  if ($in_build_tree) {
 	    info_helpfile("$conf{top_srcdir}/gasnet/README");
 	  } else {
-	    info_helpfile("$upcr_share/doc/gasnet/README");
+	    info_helpfile("$upcr_share/doc/GASNet/README");
 	  }
 	} else {
 	  $opt_help_gasnet =~ s/-conduit//;
@@ -172,7 +172,7 @@ sub getOpts
 	      info_helpfile("$conf{top_srcdir}/gasnet/".$opt_help_gasnet."-conduit/README");
 	    }
 	  } else {
-	    info_helpfile("$upcr_share/doc/gasnet/README-".$opt_help_gasnet);
+	    info_helpfile("$upcr_share/doc/GASNet/README-".$opt_help_gasnet);
 	  }
 	}
     }
@@ -510,8 +510,8 @@ sub verbose
 
 sub gasnet_encode {
     my ($in) = @_;
-    $in =~ s/%(0[0-9A-Fa-f][0-9A-Fa-f])/%025$1/g; # prevent false decodes
-    $in =~ s/([^A-Za-z0-9%_,\.\/:+=@^-])/sprintf("%%0%02x",(ord($1)))/ge;
+    $in =~ s!%(0[0-9A-Fa-f]{2})!%025$1!g; # prevent false decodes
+    $in =~ s!([^A-Za-z0-9%_,\./:+=@^-])!'%0'.unpack("H2",$1)!ge;
     return $in;
 }
 
@@ -551,7 +551,7 @@ sub initialize
     $upcr_bin = dirname($upcr_bin);    # from File::Basename
     # if we're in build tree, everything in same dir
     if (-f "$upcr_bin/upcrun.conf") {
-        $upcr_home = $upcr_etc = $upcr_lib = $upcr_include = $upcr_share = $upcr_bin;
+        $upcr_home = $upcr_etc = $upcr_include = $upcr_share = $upcr_bin;
 	$in_build_tree = 1;
     } else {
         # should be in $prefix/bin part of an installed tree, with 'include' and
@@ -559,7 +559,6 @@ sub initialize
         $upcr_bin =~ m@^(.*?)/bin$@;
         $upcr_home = $1;
         $upcr_etc = "$upcr_home/etc";
-        $upcr_lib = "$upcr_home/lib";
         $upcr_include = "$upcr_home/include";
         $upcr_share = "$upcr_home/share";
         unless (-f "$upcr_etc/upcrun.conf") {
@@ -1014,7 +1013,7 @@ sub initialize
         if (!$UPC_NODES && $ENV{PE_HOSTFILE} && -f $ENV{PE_HOSTFILE}) {
             $UPC_NODES = parseNodeFile($ENV{PE_HOSTFILE});
 	    # Only "recognized" for mpi- and ssh-spawners:
-            $recognized_file = grep($_ eq $conduit, qw(mpi mxm ibv));
+            $recognized_file = grep($_ eq $conduit, qw(mpi ofi mxm ibv));
         }
         if (!$UPC_NODES && $ENV{SLURM_NODELIST}) {
             $UPC_NODES = `scontrol show hostnames $ENV{SLURM_NODELIST}`
@@ -1516,6 +1515,7 @@ sub runDriver
     # just truncating the final proc's count ("blocked" layout).
     # We are also assuming block layout of processes by the spawner!!
     my $numproc = 0;
+    my $max_th_per_proc = 0;
     if (!isPthreads()) {
 	# No pthreads -> easy case
         $numproc = $opt_numthreads;
@@ -1527,6 +1527,7 @@ sub runDriver
             die "Value '$x' in UPC_PTHREADS_MAP is not a valid thread count.\n"
 		unless ($x =~ /^[1-9][0-9]*$/);
             $sum += $x;
+            $max_th_per_proc = max($max_th_per_proc, $x);
             $numproc++;
         }
         die "UPC_PTHREADS_MAP does not match the number of UPC threads ($opt_numthreads).\n"
@@ -1535,6 +1536,7 @@ sub runDriver
 	# Things layout uniformly
         $numproc = int($opt_numthreads / $opt_th_per_proc);
 	pushVar('UPC_PTHREADS_PER_PROC', $opt_th_per_proc);
+        $max_th_per_proc = $opt_th_per_proc;
     } else {
 	# $opt_th_per_proc fails to evenly divide one or both of $opt_numthreads or $cpus_per_node
 	# Start building a UPC_PTHREADS_MAP assuming blocked layout of procs->nodes.
@@ -1577,6 +1579,14 @@ sub runDriver
 	$map =~ s/^ //;
 	
         pushVar('UPC_PTHREADS_MAP', $map);
+        $max_th_per_proc = $opt_th_per_proc;
+    }
+
+    # This is harmless if unused
+    if ($max_th_per_proc && grep($_ eq $conduit, qw(aries gemini)) && !exists $ENV{'GASNET_DOMAIN_COUNT'}) {
+      my $pthreads_per_domain = $ENV{'GASNET_GNI_PTHREADS_PER_DOMAIN'} || 1;
+      my $domains = ($max_th_per_proc + $pthreads_per_domain - 1) / $pthreads_per_domain;
+      pushVar('GASNET_DOMAIN_COUNT', $domains);
     }
 
     if ($numnodes > $numproc) {
@@ -1669,21 +1679,21 @@ sub runDriver
     $envlist = $envlist || 'UPCRI_NON_EMPTY_KLUDGE'; # ensure envlist is non-empty
     #print "env list is $envlist\n";
 
-    # perform encoding
+    # perform encoding (NOTE: too late to use pushVar here)
     if ($opt_encode_env) {
       my @vars = split(/,/,$envlist);
       push(@vars, grep {m/^GASNET_/} (keys %ENV));
       for my $varname (@vars) {
         if (defined (my $val = $ENV{$varname})) {
 	  my $encval = gasnet_encode($val);
-	  pushVar($varname, $encval) if ($encval ne $val);
+	  $ENV{$varname} = $encval if ($encval ne $val);
 	}
       }
-      pushVar('GASNET_DISABLE_ENVDECODE', 0) if (defined $ENV{'GASNET_DISABLE_ENVDECODE'});  
+      $ENV{'GASNET_DISABLE_ENVDECODE'} = 0 if (defined $ENV{'GASNET_DISABLE_ENVDECODE'});
     }
     if ($opt_encode_args) {
       map { $_ = gasnet_encode($_); } @exeargs;
-      pushVar('GASNET_DISABLE_ARGDECODE', 0) if (defined $ENV{'GASNET_DISABLE_ARGDECODE'});  
+      $ENV{'GASNET_DISABLE_ARGDECODE'} = 0 if (defined $ENV{'GASNET_DISABLE_ARGDECODE'});
     }
 
     # Escape '%' in the potential substitution variables.
@@ -1718,10 +1728,11 @@ sub runDriver
 	                  } elsif (m/^%[Vv]$/) {
 			    ($opt_verboseexpansion?"-v":undef);
 			  } else {
-			    while (m,(###INSTALL_PREFIX###/bin/(\S+)\b),) {
+			    while ($in_build_tree && m,(\$\(UPCR_HOME\)/bin/(\S+)\b),) {
 			      # We must be running uninstalled.  Find the spawner.
 			      my $match = $1;
 			      my @dirs = (
+				"$conf{top_builddir}/gasnet/other/amudp",
 				"$conf{top_builddir}/gasnet/${conduit}-conduit/contrib",
 				"$conf{top_srcdir}/gasnet/${conduit}-conduit/contrib",
 				"$conf{top_builddir}/gasnet/${conduit}-conduit",
@@ -1735,8 +1746,9 @@ sub runDriver
 				last if (-f "${path}/${file}");
 			      }
 			      die "Unable to locate spawner '$file'" unless ($path);
-			      s,${match},${path}/${file},g;
+			      s,\Q${match},${path}/${file},g;
 			    }
+			    s/\$\(UPCR_HOME\)/$upcr_home/g;
 			    s/%[Dd]/$tmp_cwd/g; 
 			    s/%[Nn]/$numproc/g; 
 			    s/%[Mm]/$numnodes/g;
